import unittest
from typing import Optional

import pytest
from pyspark.sql import Row

from snorkel.map import Mapper, lambda_mapper
from snorkel.map.spark import make_spark_mapper
from snorkel.types import DataPoint, FieldMap


class SplitWordsMapper(Mapper):
    def __init__(
        self, name: str, text_field: str, lower_field: str, words_field: str
    ) -> None:
        super().__init__(
            name, dict(text=text_field), dict(lower=lower_field, words=words_field)
        )

    def run(self, text: str) -> FieldMap:  # type: ignore
        return dict(lower=text.lower(), words=text.split())


class SplitWordsMapperDefaultArgs(Mapper):
    def run(self, text: str) -> FieldMap:  # type: ignore
        return dict(lower=text.lower(), words=text.split())


class MapperReturnsNone(Mapper):
    def run(self, text: str) -> Optional[FieldMap]:  # type: ignore
        return None


class SquareHitTracker:
    def __init__(self):
        self.n_hits = 0

    def __call__(self, x: float) -> float:
        self.n_hits += 1
        return x**2


@lambda_mapper()
def square(x: DataPoint) -> DataPoint:
    fields = x.asDict()
    fields["num_squared"] = x.num**2
    return Row(**fields)


@lambda_mapper()
def modify_in_place(x: DataPoint) -> DataPoint:
    x.d["my_key"] = 0
    return Row(num=x.num, d=x.d, d_new=x.d)


class TestMapperCore(unittest.TestCase):
    def _get_x(self, num=8, text="Henry has fun") -> Row:
        return Row(num=num, text=text)

    def _get_x_dict(self) -> Row:
        return Row(num=8, d=dict(my_key=1))

    @pytest.mark.spark
    def test_numeric_mapper(self) -> None:
        x_mapped = square(self._get_x())
        # NB: not using `self.assertIsNotNone` due to mypy
        # See https://github.com/python/mypy/issues/5088
        assert x_mapped is not None
        self.assertEqual(x_mapped.num, 8)
        self.assertEqual(x_mapped.text, "Henry has fun")
        self.assertEqual(x_mapped.num_squared, 64)

    @pytest.mark.spark
    def test_text_mapper(self) -> None:
        split_words = SplitWordsMapper(
            "split_words", "text", "text_lower", "text_words"
        )
        split_words_spark = make_spark_mapper(split_words)
        x_mapped = split_words_spark(self._get_x())
        assert x_mapped is not None
        self.assertEqual(x_mapped.num, 8)
        self.assertEqual(x_mapped.text, "Henry has fun")
        self.assertEqual(x_mapped.text_lower, "henry has fun")
        self.assertEqual(x_mapped.text_words, ["Henry", "has", "fun"])

    @pytest.mark.spark
    def test_mapper_same_field(self) -> None:
        split_words = SplitWordsMapper("split_words", "text", "text", "text_words")
        split_words_spark = make_spark_mapper(split_words)
        x = self._get_x()
        x_mapped = split_words_spark(x)
        self.assertEqual(x.num, 8)
        self.assertEqual(x.text, "Henry has fun")
        self.assertFalse(hasattr(x, "text_words"))
        assert x_mapped is not None
        self.assertEqual(x_mapped.num, 8)
        self.assertEqual(x_mapped.text, "henry has fun")
        self.assertEqual(x_mapped.text_words, ["Henry", "has", "fun"])

    @pytest.mark.spark
    def test_mapper_default_args(self) -> None:
        split_words = SplitWordsMapperDefaultArgs("split_words")
        split_words_spark = make_spark_mapper(split_words)
        x_mapped = split_words_spark(self._get_x())
        assert x_mapped is not None
        self.assertEqual(x_mapped.num, 8)
        self.assertEqual(x_mapped.text, "Henry has fun")
        self.assertEqual(x_mapped.lower, "henry has fun")
        self.assertEqual(x_mapped.words, ["Henry", "has", "fun"])

    @pytest.mark.spark
    def test_mapper_in_place(self) -> None:
        x = self._get_x_dict()
        x_mapped = modify_in_place(x)
        self.assertEqual(x.num, 8)
        self.assertEqual(x.d, dict(my_key=1))
        self.assertFalse(hasattr(x, "d_new"))
        assert x_mapped is not None
        self.assertEqual(x_mapped.num, 8)
        self.assertEqual(x_mapped.d, dict(my_key=0))
        self.assertEqual(x_mapped.d_new, dict(my_key=0))

    @pytest.mark.spark
    def test_mapper_returns_none(self) -> None:
        mapper = MapperReturnsNone("none_mapper")
        mapper_spark = make_spark_mapper(mapper)
        x_mapped = mapper_spark(self._get_x())
        self.assertIsNone(x_mapped)

    @pytest.mark.spark
    def test_decorator_mapper_memoized(self) -> None:
        square_hit_tracker = SquareHitTracker()

        @lambda_mapper(memoize=True)
        def square(x: DataPoint) -> DataPoint:
            fields = x.asDict()
            fields["num_squared"] = square_hit_tracker(x.num)
            return Row(**fields)

        x8 = self._get_x()
        x9 = self._get_x(9)
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 1)
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 1)
        x19_mapped = square(x9)
        assert x19_mapped is not None
        self.assertEqual(x19_mapped.num_squared, 81)
        self.assertEqual(square_hit_tracker.n_hits, 2)
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 2)

        square.reset_cache()
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 3)

    @pytest.mark.spark
    def test_decorator_mapper_memoized_none(self) -> None:
        square_hit_tracker = SquareHitTracker()

        @lambda_mapper(memoize=True)
        def square(x: DataPoint) -> DataPoint:
            fields = x.asDict()
            fields["num_squared"] = square_hit_tracker(x.num)
            if x.num == 21:
                return None
            return Row(**fields)

        x21 = self._get_x(21)
        x21_mapped = square(x21)
        self.assertIsNone(x21_mapped)
        self.assertEqual(square_hit_tracker.n_hits, 1)
        x21_mapped = square(x21)
        self.assertIsNone(x21_mapped)
        self.assertEqual(square_hit_tracker.n_hits, 1)

    @pytest.mark.spark
    def test_decorator_mapper_not_memoized(self) -> None:
        square_hit_tracker = SquareHitTracker()

        @lambda_mapper(memoize=False)
        def square(x: DataPoint) -> DataPoint:
            fields = x.asDict()
            fields["num_squared"] = square_hit_tracker(x.num)
            return Row(**fields)

        x8 = self._get_x()
        x9 = self._get_x(9)
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 1)
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 2)
        x19_mapped = square(x9)
        assert x19_mapped is not None
        self.assertEqual(x19_mapped.num_squared, 81)
        self.assertEqual(square_hit_tracker.n_hits, 3)
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 4)



import unittest

import pandas as pd
import torch
import torch.nn as nn

from snorkel.classification import DictDataLoader, DictDataset, Operation, Task
from snorkel.slicing import (
    PandasSFApplier,
    add_slice_labels,
    convert_to_slice_tasks,
    slicing_function,
)


@slicing_function()
def f(x):
    return x.val < 0.25


class UtilsTest(unittest.TestCase):
    def test_add_slice_labels(self):
        # Create dummy data
        # Given slicing function f(), we expect the first two entries to be active
        x = torch.Tensor([0.1, 0.2, 0.3, 0.4, 0.5])
        y = torch.Tensor([0, 1, 1, 0, 1]).long()
        dataset = DictDataset(
            name="TestData", split="train", X_dict={"data": x}, Y_dict={"TestTask": y}
        )

        # Ensure that we start with 1 labelset
        self.assertEqual(len(dataset.Y_dict), 1)

        # Apply SFs with PandasSFApplier
        df = pd.DataFrame({"val": x, "y": y})
        slicing_functions = [f]
        applier = PandasSFApplier(slicing_functions)
        S = applier.apply(df, progress_bar=False)

        dataloader = DictDataLoader(dataset)

        dummy_task = create_dummy_task(task_name="TestTask")
        add_slice_labels(dataloader, dummy_task, S)

        # Ensure that all the fields are present
        labelsets = dataloader.dataset.Y_dict
        self.assertIn("TestTask", labelsets)
        self.assertIn("TestTask_slice:base_ind", labelsets)
        self.assertIn("TestTask_slice:base_pred", labelsets)
        self.assertIn("TestTask_slice:f_ind", labelsets)
        self.assertIn("TestTask_slice:f_pred", labelsets)
        self.assertEqual(len(labelsets), 5)

        # Ensure "ind" contains mask
        self.assertEqual(
            labelsets["TestTask_slice:f_ind"].numpy().tolist(), [1, 1, 0, 0, 0]
        )
        self.assertEqual(
            labelsets["TestTask_slice:base_ind"].numpy().tolist(), [1, 1, 1, 1, 1]
        )

        # Ensure "pred" contains masked elements
        self.assertEqual(
            labelsets["TestTask_slice:f_pred"].numpy().tolist(), [0, 1, -1, -1, -1]
        )
        self.assertEqual(
            labelsets["TestTask_slice:base_pred"].numpy().tolist(), [0, 1, 1, 0, 1]
        )
        self.assertEqual(labelsets["TestTask"].numpy().tolist(), [0, 1, 1, 0, 1])

    def test_convert_to_slice_tasks(self):
        task_name = "TestTask"
        task = create_dummy_task(task_name)

        slice_names = ["slice_a", "slice_b", "slice_c"]
        slice_tasks = convert_to_slice_tasks(task, slice_names)

        slice_task_names = [t.name for t in slice_tasks]
        # Check for original base task
        self.assertIn(task_name, slice_task_names)

        # Check for 2 tasks (pred + ind) per slice, accounting for base slice
        for slice_name in slice_names + ["base"]:
            self.assertIn(f"{task_name}_slice:{slice_name}_pred", slice_task_names)
            self.assertIn(f"{task_name}_slice:{slice_name}_ind", slice_task_names)

        self.assertEqual(len(slice_tasks), 2 * (len(slice_names) + 1) + 1)

        # Test that modules share the same body flow operations
        # NOTE: Use "is" comparison to check object equality
        body_flow = task.op_sequence[:-1]
        ind_and_pred_tasks = [
            t for t in slice_tasks if "_ind" in t.name or "_pred" in t.name
        ]
        for op in body_flow:
            for slice_task in ind_and_pred_tasks:
                self.assertTrue(
                    slice_task.module_pool[op.module_name]
                    is task.module_pool[op.module_name]
                )

        # Test that pred tasks share the same predictor head
        pred_tasks = [t for t in slice_tasks if "_pred" in t.name]
        predictor_head_name = pred_tasks[0].op_sequence[-1].module_name
        shared_predictor_head = pred_tasks[0].module_pool[predictor_head_name]
        for pred_task in pred_tasks[1:]:
            self.assertTrue(
                pred_task.module_pool[predictor_head_name] is shared_predictor_head
            )


def create_dummy_task(task_name):
    # Create dummy task
    module_pool = nn.ModuleDict(
        {"linear1": nn.Linear(2, 10), "linear2": nn.Linear(10, 2)}
    )

    op_sequence = [
        Operation(name="encoder", module_name="linear1", inputs=["_input_"]),
        Operation(name="prediction_head", module_name="linear2", inputs=["encoder"]),
    ]

    task = Task(name=task_name, module_pool=module_pool, op_sequence=op_sequence)
    return task



import unittest
from types import SimpleNamespace

from snorkel.preprocess import preprocessor
from snorkel.slicing.sf.nlp import NLPSlicingFunction, nlp_slicing_function
from snorkel.types import DataPoint


@preprocessor()
def combine_text(x: DataPoint) -> DataPoint:
    x.text = f"{x.title} {x.article}"
    return x


def has_person_mention(x: DataPoint) -> int:
    person_ents = [ent for ent in x.doc.ents if ent.label_ == "PERSON"]
    return 0 if len(person_ents) > 0 else -1


class TestNLPSlicingFunction(unittest.TestCase):
    def _run_sf(self, sf: NLPSlicingFunction) -> None:
        x = SimpleNamespace(
            num=8, title="Great film!", article="The movie is really great!"
        )
        self.assertEqual(sf(x), -1)
        x = SimpleNamespace(num=8, title="Nice movie!", article="Jane Doe acted well.")
        self.assertEqual(sf(x), 0)

    def test_nlp_slicing_function(self) -> None:
        sf = NLPSlicingFunction(name="my_sf", f=has_person_mention, pre=[combine_text])
        self._run_sf(sf)

    def test_nlp_slicing_function_decorator(self) -> None:
        @nlp_slicing_function(pre=[combine_text])
        def has_person_mention(x: DataPoint) -> int:
            person_ents = [ent for ent in x.doc.ents if ent.label_ == "PERSON"]
            return 0 if len(person_ents) > 0 else -1

        self.assertIsInstance(has_person_mention, NLPSlicingFunction)
        self.assertEqual(has_person_mention.name, "has_person_mention")
        self._run_sf(has_person_mention)



import unittest

from snorkel.labeling.model.logger import Logger


class LoggerTest(unittest.TestCase):
    def test_basic(self):
        metrics_dict = {"train/loss": 0.01}
        logger = Logger(log_freq=1)
        logger.log(metrics_dict)

        metrics_dict = {"train/message": "well done!"}
        logger = Logger(log_freq=1)
        logger.log(metrics_dict)

    def test_bad_metrics_dict(self):
        bad_metrics_dict = {"task1/slice1/train/loss": 0.05}
        logger = Logger(log_freq=1)
        self.assertRaises(Exception, logger.log, bad_metrics_dict)

    def test_valid_metrics_dict(self):
        mtl_metrics_dict = {"task1/valid/loss": 0.05}
        logger = Logger(log_freq=1)
        logger.log(mtl_metrics_dict)


if __name__ == "__main__":
    unittest.main()



import unittest

from snorkel.augmentation import ApplyAllPolicy, ApplyEachPolicy


class TestPolicy(unittest.TestCase):
    def test_apply_each_policy(self):
        policy = ApplyEachPolicy(3, keep_original=True)
        samples = policy.generate_for_example()
        self.assertEqual(samples, [[], [0], [1], [2]])

        policy = ApplyEachPolicy(3, keep_original=False)
        samples = policy.generate_for_example()
        self.assertEqual(samples, [[0], [1], [2]])

    def test_apply_all_policy(self):
        policy = ApplyAllPolicy(3, n_per_original=2, keep_original=False)
        samples = policy.generate_for_example()
        self.assertEqual(samples, [[0, 1, 2], [0, 1, 2]])



from pyspark.sql import Row

from snorkel.types import FieldMap

from .core import Mapper


def _update_fields(x: Row, mapped_fields: FieldMap) -> Row:
    # ``pyspark.sql.Row`` objects are not mutable, so need to
    # reconstruct
    all_fields = x.asDict()
    all_fields.update(mapped_fields)
    return Row(**all_fields)


def make_spark_mapper(mapper: Mapper) -> Mapper:
    """Convert ``Mapper`` to be compatible with PySpark.

    Parameters
    ----------
    mapper
        Mapper to make compatible with PySpark
    """
    mapper._update_fields = _update_fields  # type: ignore
    return mapper



import unittest

from snorkel.types import Config
from snorkel.utils.config_utils import merge_config


class FooConfig(Config):
    a: float = 0.5


class BarConfig(Config):
    a: int = 1
    foo_config: FooConfig = FooConfig()  # type: ignore


class UtilsTest(unittest.TestCase):
    def test_merge_config(self):
        config_updates = {"a": 2, "foo_config": {"a": 0.75}}
        bar_config = merge_config(BarConfig(), config_updates)
        self.assertEqual(bar_config.a, 2)
        self.assertEqual(bar_config.foo_config.a, 0.75)



import unittest

import numpy as np
import pandas as pd

from snorkel.labeling import filter_unlabeled_dataframe


class TestAnalysis(unittest.TestCase):
    def test_filter_unlabeled_dataframe(self) -> None:
        X = pd.DataFrame(dict(A=["x", "y", "z"], B=[1, 2, 3]))
        y = np.array(
            [[0.25, 0.25, 0.25, 0.25], [1.0, 0.0, 0.0, 0.0], [0.2, 0.3, 0.5, 0.0]]
        )
        L = np.array([[0, 1, -1], [-1, -1, -1], [1, 1, 0]])
        X_filtered, y_filtered = filter_unlabeled_dataframe(X, y, L)
        np.array_equal(X_filtered.values, np.array([["x", 1], ["z", 3]]))
        np.testing.assert_array_almost_equal(
            y_filtered, np.array([[0.25, 0.25, 0.25, 0.25], [0.2, 0.3, 0.5, 0.0]])
        )



"""
This script is used to manually test
    `snorkel.labeling.apply.lf_applier_spark.SparkLFApplier`

To test on AWS EMR:
    1. Allocate an EMR cluster (e.g. label 5.24.0) with > 1 worker and SSH permissions
    2. Clone and pip install snorkel on the master node
        ```
        sudo yum install git
        git clone https://github.com/snorkel-team/snorkel
        cd snorkel
        python3 -m pip install -t snorkel-package .
        cd snorkel-package
        zip -r ../snorkel-package.zip .
        cd ..
        ```
    3. Run
        ```
        sudo sed -i -e \
            '$a\\export PYSPARK_PYTHON=/usr/bin/python3' \
            /etc/spark/conf/spark-env.sh
        ```
    4. Run
        ```
        spark-submit \
            --py-files snorkel-package.zip \
            test/labeling/apply/lf_applier_spark_test_script.py
        ```
"""

import logging
from typing import List

import numpy as np
from pyspark import SparkContext

from snorkel.labeling.apply.spark import SparkLFApplier
from snorkel.labeling.lf import labeling_function
from snorkel.types import DataPoint

logging.basicConfig(level=logging.INFO)


@labeling_function()
def f(x: DataPoint) -> int:
    return 1 if x > 42 else 0


@labeling_function(resources=dict(db=[3, 6, 9]))
def g(x: DataPoint, db: List[int]) -> int:
    return 1 if x in db else 0


DATA = [3, 43, 12, 9]
L_EXPECTED = np.array([[0, 1], [1, 0], [0, 0], [0, 1]])


def build_lf_matrix() -> None:
    logging.info("Getting Spark context")
    sc = SparkContext()
    sc.addPyFile("snorkel-package.zip")
    rdd = sc.parallelize(DATA)

    logging.info("Applying LFs")
    lf_applier = SparkLFApplier([f, g])
    L = lf_applier.apply(rdd)

    np.testing.assert_equal(L.toarray(), L_EXPECTED)


if __name__ == "__main__":
    build_lf_matrix()



import unittest
from typing import Optional

import numpy as np

from snorkel.labeling import LFAnalysis
from snorkel.synthetic.synthetic_data import generate_simple_label_matrix


class TestGenerateSimpleLabelMatrix(unittest.TestCase):
    """Testing the generate_simple_label_matrix function."""

    def setUp(self) -> None:
        """Set constants for the tests."""
        self.m = 10  # Number of LFs
        self.n = 1000  # Number of data points

    def _test_generate_L(self, k: int, decimal: Optional[int] = 2) -> None:
        """Test generated label matrix L for consistency with P, Y.

        This tests for consistency between the true conditional LF probabilities, P,
        and the empirical ones computed from L and Y, where P, L, and Y are generated
        by the generate_simple_label_matrix function.

        Parameters
        ----------
        k
            Cardinality
        decimal
            Number of decimals to check element-wise error, err < 1.5 * 10**(-decimal)
        """
        np.random.seed(123)
        P, Y, L = generate_simple_label_matrix(self.n, self.m, k)
        P_emp = LFAnalysis(L).lf_empirical_probs(Y, k=k)
        np.testing.assert_array_almost_equal(P, P_emp, decimal=decimal)

    def test_generate_L(self) -> None:
        """Test the generated dataset for consistency."""
        self._test_generate_L(2, decimal=1)

    def test_generate_L_multiclass(self) -> None:
        """Test the generated dataset for consistency with cardinality=3."""
        self._test_generate_L(3, decimal=1)


if __name__ == "__main__":
    unittest.main()



import unittest

from snorkel.utils.data_operators import check_unique_names


class DataOperatorsTest(unittest.TestCase):
    def test_check_unique_names(self):
        check_unique_names(["alice", "bob", "chuck"])
        with self.assertRaisesRegex(ValueError, "3 operators with name c"):
            check_unique_names(["a", "a", "b", "c", "c", "c"])



import os
import random
import tempfile
import unittest

import numpy as np
import pandas as pd
import torch
import torch.nn as nn

from snorkel.classification import (
    DictDataLoader,
    DictDataset,
    MultitaskClassifier,
    Operation,
    Task,
)

NUM_EXAMPLES = 10
BATCH_SIZE = 2


class ClassifierTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.task1 = create_task("task1", module_suffixes=["A", "A"])
        cls.task2 = create_task("task2", module_suffixes=["B", "B"])
        cls.dataloader = create_dataloader("task1")

    def setUp(self):
        random.seed(123)
        np.random.seed(123)
        torch.manual_seed(123)

    def test_onetask_model(self):
        model = MultitaskClassifier(tasks=[self.task1])
        self.assertEqual(len(model.task_names), 1)
        self.assertEqual(len(model.op_sequences), 1)
        self.assertEqual(len(model.module_pool), 2)

    def test_twotask_none_overlap_model(self):
        """Add two tasks with totally separate modules and flows"""
        model = MultitaskClassifier(tasks=[self.task1, self.task2])
        self.assertEqual(len(model.task_names), 2)
        self.assertEqual(len(model.op_sequences), 2)
        self.assertEqual(len(model.module_pool), 4)

    def test_twotask_all_overlap_model(self):
        """Add two tasks with identical modules and flows"""
        task1 = create_task("task1", module_suffixes=["A", "A"])
        task2 = create_task("task2", module_suffixes=["A", "A"])
        model = MultitaskClassifier(tasks=[task1, task2])
        self.assertEqual(len(model.task_names), 2)
        self.assertEqual(len(model.op_sequences), 2)
        self.assertEqual(len(model.module_pool), 2)

    def test_twotask_partial_overlap_model(self):
        """Add two tasks with overlapping modules and flows"""
        task1 = create_task("task1", module_suffixes=["A", "A"])
        task2 = create_task("task2", module_suffixes=["A", "B"])
        model = MultitaskClassifier(tasks=[task1, task2])
        self.assertEqual(len(model.task_names), 2)
        self.assertEqual(len(model.op_sequences), 2)
        self.assertEqual(len(model.module_pool), 3)

    def test_bad_tasks(self):
        with self.assertRaisesRegex(ValueError, "Found duplicate task"):
            MultitaskClassifier(tasks=[self.task1, self.task1])
        with self.assertRaisesRegex(ValueError, "Unrecognized task type"):
            MultitaskClassifier(tasks=[self.task1, {"fake_task": 42}])
        with self.assertRaisesRegex(ValueError, "Unsuccessful operation"):
            task1 = create_task("task1")
            task1.op_sequence[0].inputs[0] = (0, 0)
            model = MultitaskClassifier(tasks=[task1])
            X_dict = self.dataloader.dataset.X_dict
            model.forward(X_dict, [task1.name])

    def test_no_data_parallel(self):
        model = MultitaskClassifier(tasks=[self.task1, self.task2], dataparallel=False)
        self.assertEqual(len(model.task_names), 2)
        self.assertIsInstance(model.module_pool["linear1A"], nn.Module)

    def test_no_input_spec(self):
        # Confirm model doesn't break when a module does not specify specific inputs
        dataset = create_dataloader("task", shuffle=False).dataset
        task = Task(
            name="task",
            module_pool=nn.ModuleDict({"identity": nn.Identity()}),
            op_sequence=[Operation("identity", [])],
        )
        model = MultitaskClassifier(tasks=[task], dataparallel=False)
        outputs = model.forward(dataset.X_dict, ["task"])
        self.assertIn("_input_", outputs)

    def test_predict(self):
        model = MultitaskClassifier([self.task1])
        results = model.predict(self.dataloader)
        self.assertEqual(sorted(list(results.keys())), ["golds", "probs"])
        np.testing.assert_array_equal(
            results["golds"]["task1"], self.dataloader.dataset.Y_dict["task1"].numpy()
        )
        np.testing.assert_array_equal(
            results["probs"]["task1"], np.ones((NUM_EXAMPLES, 2)) * 0.5
        )

        results = model.predict(self.dataloader, return_preds=True)
        self.assertEqual(sorted(list(results.keys())), ["golds", "preds", "probs"])
        # deterministic random tie breaking alternates predicted labels
        np.testing.assert_array_equal(
            results["preds"]["task1"],
            np.array([0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0]),
        )

    def test_empty_batch(self):
        dataset = create_dataloader("task1", shuffle=False).dataset
        dataset.Y_dict["task1"] = torch.full_like(dataset.Y_dict["task1"], -1)
        model = MultitaskClassifier([self.task1])
        loss_dict, count_dict = model.calculate_loss(dataset.X_dict, dataset.Y_dict)
        self.assertFalse(loss_dict)
        self.assertFalse(count_dict)

    def test_partially_empty_batch(self):
        dataset = create_dataloader("task1", shuffle=False).dataset
        dataset.Y_dict["task1"][0] = -1
        model = MultitaskClassifier([self.task1])
        loss_dict, count_dict = model.calculate_loss(dataset.X_dict, dataset.Y_dict)
        self.assertEqual(count_dict["task1"], 9)

    def test_remapped_labels(self):
        # Test additional label keys in the Y_dict
        # Without remapping, model should ignore them
        task_name = self.task1.name
        X = torch.FloatTensor([[i, i] for i in range(NUM_EXAMPLES)])
        Y = torch.ones(NUM_EXAMPLES).long()

        Y_dict = {task_name: Y, "other_task": Y}
        dataset = DictDataset(
            name="dataset", split="train", X_dict={"data": X}, Y_dict=Y_dict
        )
        dataloader = DictDataLoader(dataset, batch_size=BATCH_SIZE)

        model = MultitaskClassifier([self.task1])
        loss_dict, count_dict = model.calculate_loss(dataset.X_dict, dataset.Y_dict)
        self.assertIn("task1", loss_dict)

        # Test setting without remapping
        results = model.predict(dataloader)
        self.assertIn("task1", results["golds"])
        self.assertNotIn("other_task", results["golds"])
        scores = model.score([dataloader])
        self.assertIn("task1/dataset/train/accuracy", scores)
        self.assertNotIn("other_task/dataset/train/accuracy", scores)

        # Test remapped labelsets
        results = model.predict(dataloader, remap_labels={"other_task": task_name})
        self.assertIn("task1", results["golds"])
        self.assertIn("other_task", results["golds"])
        results = model.score([dataloader], remap_labels={"other_task": task_name})
        self.assertIn("task1/dataset/train/accuracy", results)
        self.assertIn("other_task/dataset/train/accuracy", results)

    def test_score(self):
        model = MultitaskClassifier([self.task1])
        metrics = model.score([self.dataloader])
        # deterministic random tie breaking alternates predicted labels
        self.assertEqual(metrics["task1/dataset/train/accuracy"], 0.4)

        # test dataframe format
        metrics_df = model.score([self.dataloader], as_dataframe=True)
        self.assertTrue(isinstance(metrics_df, pd.DataFrame))
        self.assertEqual(metrics_df.at[0, "score"], 0.4)

    def test_score_shuffled(self):
        # Test scoring with a shuffled dataset

        class SimpleVoter(nn.Module):
            def forward(self, x):
                """Set class 0 to -1 if x and 1 otherwise"""
                mask = x % 2 == 0
                out = torch.zeros(x.shape[0], 2)
                out[mask, 0] = 1  # class 0
                out[~mask, 1] = 1  # class 1
                return out

        # Create model
        task_name = "VotingTask"
        module_name = "simple_voter"
        module_pool = nn.ModuleDict({module_name: SimpleVoter()})
        op0 = Operation(
            module_name=module_name, inputs=[("_input_", "data")], name="op0"
        )
        op_sequence = [op0]
        task = Task(name=task_name, module_pool=module_pool, op_sequence=op_sequence)
        model = MultitaskClassifier([task])

        # Create dataset
        y_list = [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]
        x_list = [i for i in range(len(y_list))]
        Y = torch.LongTensor(y_list * 100)
        X = torch.FloatTensor(x_list * 100)
        dataset = DictDataset(
            name="dataset", split="train", X_dict={"data": X}, Y_dict={task_name: Y}
        )

        # Create dataloaders
        dataloader = DictDataLoader(dataset, batch_size=2, shuffle=False)
        scores = model.score([dataloader])

        self.assertEqual(scores["VotingTask/dataset/train/accuracy"], 0.6)

        dataloader_shuffled = DictDataLoader(dataset, batch_size=2, shuffle=True)
        scores_shuffled = model.score([dataloader_shuffled])
        self.assertEqual(scores_shuffled["VotingTask/dataset/train/accuracy"], 0.6)

    def test_save_load(self):
        fd, checkpoint_path = tempfile.mkstemp()

        task1 = create_task("task1")
        task2 = create_task("task2")
        # Make task2's second linear layer have different weights
        task2.module_pool["linear2"] = nn.Linear(2, 2)

        model = MultitaskClassifier([task1])
        self.assertTrue(
            torch.eq(
                task1.module_pool["linear2"].weight,
                model.module_pool["linear2"].module.weight,
            ).all()
        )
        model.save(checkpoint_path)
        model = MultitaskClassifier([task2])
        self.assertFalse(
            torch.eq(
                task1.module_pool["linear2"].weight,
                model.module_pool["linear2"].module.weight,
            ).all()
        )
        model.load(checkpoint_path)
        self.assertTrue(
            torch.eq(
                task1.module_pool["linear2"].weight,
                model.module_pool["linear2"].module.weight,
            ).all()
        )

        os.close(fd)


def create_dataloader(task_name="task", split="train", **kwargs):
    X = torch.FloatTensor([[i, i] for i in range(NUM_EXAMPLES)])
    Y = torch.ones(NUM_EXAMPLES).long()

    dataset = DictDataset(
        name="dataset", split=split, X_dict={"data": X}, Y_dict={task_name: Y}
    )

    dataloader = DictDataLoader(dataset, batch_size=BATCH_SIZE, **kwargs)
    return dataloader


def create_task(task_name, module_suffixes=("", "")):
    module1_name = f"linear1{module_suffixes[0]}"
    module2_name = f"linear2{module_suffixes[1]}"

    linear1 = nn.Linear(2, 2)
    linear1.weight.data.copy_(torch.eye(2))
    linear1.bias.data.copy_(torch.zeros((2,)))

    linear2 = nn.Linear(2, 2)
    linear2.weight.data.copy_(torch.eye(2))
    linear2.bias.data.copy_(torch.zeros((2,)))

    module_pool = nn.ModuleDict(
        {module1_name: nn.Sequential(linear1, nn.ReLU()), module2_name: linear2}
    )

    op0 = Operation(module_name=module1_name, inputs=[("_input_", "data")], name="op0")
    op1 = Operation(module_name=module2_name, inputs=[op0.name], name="op1")

    op_sequence = [op0, op1]

    task = Task(name=task_name, module_pool=module_pool, op_sequence=op_sequence)

    return task


if __name__ == "__main__":
    unittest.main()



import random
import unittest

import numpy as np
import pandas as pd
import pytest
import torch

from snorkel.labeling import LabelingFunction, PandasLFApplier, labeling_function
from snorkel.labeling.model import LabelModel
from snorkel.preprocess import preprocessor
from snorkel.types import DataPoint


def create_data(n: int) -> pd.DataFrame:
    """Create random pairs x1, x2 in [-1., 1.] with label x1 > x2 + 0.25."""
    X = np.random.random((n, 2)) * 2 - 1
    Y = (X[:, 0] > X[:, 1] + 0.25).astype(int)

    df = pd.DataFrame(
        {"x0": np.random.randint(0, 1000, n), "x1": X[:, 0], "x2": X[:, 1], "y": Y}
    )
    return df


def get_positive_labeling_function(divisor: int) -> LabelingFunction:
    """Get LabelingFunction that abstains unless x0 is divisible by divisor."""

    def f(x):
        return 1 if x.x0 % divisor == 0 and x.x1 > x.x2 + 0.25 else -1

    return LabelingFunction(f"lf_pos_{divisor}", f)


def get_negative_labeling_function(divisor: int) -> LabelingFunction:
    """Get LabelingFunction that abstains unless x0 is divisible by divisor."""

    def f(x):
        return 0 if x.x0 % divisor == 0 and x.x1 <= x.x2 + 0.25 else -1

    return LabelingFunction(f"lf_neg_{divisor}", f)


@preprocessor()
def copy_features(x: DataPoint) -> DataPoint:
    """Compute x2 + 0.25 for direct comparison to x1."""
    x.x3 = x.x2 + 0.25
    return x


@labeling_function(pre=[copy_features], resources=dict(divisor=3))
def f(x: DataPoint, divisor: int) -> int:
    # Abstain unless x0 is divisible by divisor.
    return 0 if x.x0 % divisor == 1 and x.x1 > x.x3 else -1


class LabelingConvergenceTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Ensure deterministic runs
        random.seed(123)
        np.random.seed(123)
        torch.manual_seed(123)

        # Create raw data
        cls.N_TRAIN = 1500

        cls.cardinality = 2
        cls.df_train = create_data(cls.N_TRAIN)

    @pytest.mark.complex
    def test_labeling_convergence(self) -> None:
        """Test convergence of end to end labeling pipeline."""
        # Apply LFs
        labeling_functions = (
            [f]
            + [get_positive_labeling_function(divisor) for divisor in range(2, 9)]
            + [get_negative_labeling_function(divisor) for divisor in range(2, 9)]
        )
        applier = PandasLFApplier(labeling_functions)
        L_train = applier.apply(self.df_train, progress_bar=False)

        self.assertEqual(L_train.shape, (self.N_TRAIN, len(labeling_functions)))

        # Train LabelModel
        label_model = LabelModel(cardinality=self.cardinality, verbose=False)
        label_model.fit(L_train, n_epochs=100, lr=0.01, l2=0.0)
        Y_lm = label_model.predict_proba(L_train).argmax(axis=1)
        Y = self.df_train.y
        err = np.where(Y != Y_lm, 1, 0).sum() / self.N_TRAIN
        self.assertLess(err, 0.06)


if __name__ == "__main__":
    unittest.main()



import unittest

import torch.nn as nn

from snorkel.classification import Operation, Task

TASK_NAME = "TestTask"


class TaskTest(unittest.TestCase):
    def test_task_creation(self):
        module_pool = nn.ModuleDict(
            {
                "linear1": nn.Sequential(nn.Linear(2, 10), nn.ReLU()),
                "linear2": nn.Linear(10, 1),
            }
        )

        op_sequence = [
            Operation(
                name="the_first_layer", module_name="linear1", inputs=["_input_"]
            ),
            Operation(
                name="the_second_layer",
                module_name="linear2",
                inputs=["the_first_layer"],
            ),
        ]

        task = Task(name=TASK_NAME, module_pool=module_pool, op_sequence=op_sequence)

        # Task has no functionality on its own
        # Here we only confirm that the object was initialized
        self.assertEqual(task.name, TASK_NAME)


if __name__ == "__main__":
    unittest.main()



import unittest

from snorkel.augmentation import MeanFieldPolicy, RandomPolicy


class TestSamplingPolicy(unittest.TestCase):
    def test_random_policy(self):
        policy = RandomPolicy(2, sequence_length=2)
        n_samples = 100
        samples = [policy.generate() for _ in range(n_samples)]
        a_ct = samples.count([0, 0])
        b_ct = samples.count([0, 1])
        c_ct = samples.count([1, 0])
        d_ct = samples.count([1, 1])
        self.assertGreater(a_ct, 0)
        self.assertGreater(b_ct, 0)
        self.assertGreater(c_ct, 0)
        self.assertGreater(d_ct, 0)
        self.assertEqual(a_ct + b_ct + c_ct + d_ct, n_samples)

    def test_mean_field_policy(self):
        policy = MeanFieldPolicy(2, sequence_length=2, p=[1, 0])
        n_samples = 100
        samples = [policy.generate() for _ in range(n_samples)]
        self.assertEqual(samples.count([0, 0]), n_samples)



import unittest
from types import SimpleNamespace
from typing import List

import numpy as np
import pandas as pd
import pytest
from dask import dataframe as dd

from snorkel.labeling import LFApplier, PandasLFApplier, labeling_function
from snorkel.labeling.apply.core import ApplierMetadata
from snorkel.labeling.apply.dask import DaskLFApplier, PandasParallelLFApplier
from snorkel.preprocess import preprocessor
from snorkel.preprocess.nlp import SpacyPreprocessor
from snorkel.types import DataPoint


@preprocessor()
def square(x: DataPoint) -> DataPoint:
    x.num_squared = x.num**2
    return x


class SquareHitTracker:
    def __init__(self):
        self.n_hits = 0

    def __call__(self, x: float) -> float:
        self.n_hits += 1
        return x**2


@labeling_function()
def f(x: DataPoint) -> int:
    return 0 if x.num > 42 else -1


@labeling_function(pre=[square])
def fp(x: DataPoint) -> int:
    return 0 if x.num_squared > 42 else -1


@labeling_function(resources=dict(db=[3, 6, 9]))
def g(x: DataPoint, db: List[int]) -> int:
    return 0 if x.num in db else -1


@labeling_function()
def h(x: DataPoint) -> int:
    return -1


@labeling_function()
def f_np(x: DataPoint) -> int:
    return 0 if x[1] > 42 else -1


@labeling_function(resources=dict(db=[3, 6, 9]))
def g_np(x: DataPoint, db: List[int]) -> int:
    return 0 if x[1] in db else -1


@labeling_function()
def f_bad(x: DataPoint) -> int:
    return 0 if x.mum > 42 else -1


DATA = [3, 43, 12, 9, 3]
L_EXPECTED = np.array([[-1, 0], [0, -1], [-1, -1], [-1, 0], [-1, 0]])
L_EXPECTED_BAD = np.array([[-1, -1], [0, -1], [-1, -1], [-1, -1], [-1, -1]])
L_PREPROCESS_EXPECTED = np.array([[-1, -1], [0, 0], [-1, 0], [-1, 0], [-1, -1]])

TEXT_DATA = ["Jane Doe.", "Jane plays soccer.", "Jane plays soccer."]
L_TEXT_EXPECTED = np.array([[0, -1], [0, 0], [0, 0]])


class TestLFApplier(unittest.TestCase):
    def test_lf_applier(self) -> None:
        data_points = [SimpleNamespace(num=num) for num in DATA]
        applier = LFApplier([f, g])
        L = applier.apply(data_points, progress_bar=False)
        np.testing.assert_equal(L, L_EXPECTED)
        L = applier.apply(data_points, progress_bar=True)
        np.testing.assert_equal(L, L_EXPECTED)
        L, meta = applier.apply(data_points, return_meta=True)
        np.testing.assert_equal(L, L_EXPECTED)
        self.assertEqual(meta, ApplierMetadata(dict()))

    def test_lf_applier_fault(self) -> None:
        data_points = [SimpleNamespace(num=num) for num in DATA]
        applier = LFApplier([f, f_bad])
        with self.assertRaises(AttributeError):
            applier.apply(data_points, progress_bar=False)
        L = applier.apply(data_points, progress_bar=False, fault_tolerant=True)
        np.testing.assert_equal(L, L_EXPECTED_BAD)
        L, meta = applier.apply(
            data_points, progress_bar=False, fault_tolerant=True, return_meta=True
        )
        np.testing.assert_equal(L, L_EXPECTED_BAD)
        self.assertEqual(meta, ApplierMetadata(dict(f_bad=5)))

    def test_lf_applier_preprocessor(self) -> None:
        data_points = [SimpleNamespace(num=num) for num in DATA]
        applier = LFApplier([f, fp])
        L = applier.apply(data_points, progress_bar=False)
        np.testing.assert_equal(L, L_PREPROCESS_EXPECTED)

    def test_lf_applier_preprocessor_memoized(self) -> None:
        data_points = [SimpleNamespace(num=num) for num in DATA]
        square_hit_tracker = SquareHitTracker()

        @preprocessor(memoize=True)
        def square_memoize(x: DataPoint) -> DataPoint:
            x.num_squared = square_hit_tracker(x.num)
            return x

        @labeling_function(pre=[square_memoize])
        def fp_memoized(x: DataPoint) -> int:
            return 0 if x.num_squared > 42 else -1

        applier = LFApplier([f, fp_memoized])
        L = applier.apply(data_points, progress_bar=False)
        np.testing.assert_equal(L, L_PREPROCESS_EXPECTED)
        self.assertEqual(square_hit_tracker.n_hits, 4)

    def test_lf_applier_no_labels(self) -> None:
        data_points = [SimpleNamespace(num=num) for num in DATA]
        applier = LFApplier([h])
        L = applier.apply(data_points, progress_bar=False)
        np.testing.assert_equal(L, -1)

    def test_lf_applier_numpy(self) -> None:
        X = np.vstack((DATA, DATA)).T
        applier = LFApplier([f_np, g_np])
        L = applier.apply(X, progress_bar=False)
        np.testing.assert_equal(L, L_EXPECTED)


class TestPandasApplier(unittest.TestCase):
    def test_lf_applier_pandas(self) -> None:
        df = pd.DataFrame(dict(num=DATA))
        applier = PandasLFApplier([f, g])
        L = applier.apply(df, progress_bar=False)
        np.testing.assert_equal(L, L_EXPECTED)
        L = applier.apply(df, progress_bar=True)
        np.testing.assert_equal(L, L_EXPECTED)
        L, meta = applier.apply(df, return_meta=True)
        np.testing.assert_equal(L, L_EXPECTED)
        self.assertEqual(meta, ApplierMetadata(dict()))

    def test_lf_applier_pandas_fault(self) -> None:
        df = pd.DataFrame(dict(num=DATA))
        applier = PandasLFApplier([f, f_bad])
        with self.assertRaises(AttributeError):
            applier.apply(df, progress_bar=False)
        L = applier.apply(df, progress_bar=False, fault_tolerant=True)
        np.testing.assert_equal(L, L_EXPECTED_BAD)
        L, meta = applier.apply(
            df, progress_bar=False, fault_tolerant=True, return_meta=True
        )
        np.testing.assert_equal(L, L_EXPECTED_BAD)
        self.assertEqual(meta, ApplierMetadata(dict(f_bad=5)))

    def test_lf_applier_pandas_preprocessor(self) -> None:
        df = pd.DataFrame(dict(num=DATA))
        applier = PandasLFApplier([f, fp])
        L = applier.apply(df, progress_bar=False)
        np.testing.assert_equal(L, L_PREPROCESS_EXPECTED)

    def test_lf_applier_pandas_preprocessor_memoized(self) -> None:
        square_hit_tracker = SquareHitTracker()

        @preprocessor(memoize=True)
        def square_memoize(x: DataPoint) -> DataPoint:
            x.num_squared = square_hit_tracker(x.num)
            return x

        @labeling_function(pre=[square_memoize])
        def fp_memoized(x: DataPoint) -> int:
            return 0 if x.num_squared > 42 else -1

        df = pd.DataFrame(dict(num=DATA))
        applier = PandasLFApplier([f, fp_memoized])
        L = applier.apply(df, progress_bar=False)
        np.testing.assert_equal(L, L_PREPROCESS_EXPECTED)
        self.assertEqual(square_hit_tracker.n_hits, 4)

    def test_lf_applier_pandas_spacy_preprocessor(self) -> None:
        spacy = SpacyPreprocessor(text_field="text", doc_field="doc")

        @labeling_function(pre=[spacy])
        def first_is_name(x: DataPoint) -> int:
            return 0 if x.doc[0].pos_ == "PROPN" else -1

        @labeling_function(pre=[spacy])
        def has_verb(x: DataPoint) -> int:
            return 0 if sum(t.pos_ == "VERB" for t in x.doc) > 0 else -1

        df = pd.DataFrame(dict(text=TEXT_DATA))
        applier = PandasLFApplier([first_is_name, has_verb])
        L = applier.apply(df, progress_bar=False)
        np.testing.assert_equal(L, L_TEXT_EXPECTED)

    def test_lf_applier_pandas_spacy_preprocessor_memoized(self) -> None:
        spacy = SpacyPreprocessor(text_field="text", doc_field="doc")
        spacy.memoize = True

        @labeling_function(pre=[spacy])
        def first_is_name(x: DataPoint) -> int:
            return 0 if x.doc[0].pos_ == "PROPN" else -1

        @labeling_function(pre=[spacy])
        def has_verb(x: DataPoint) -> int:
            return 0 if sum(t.pos_ == "VERB" for t in x.doc) > 0 else -1

        df = pd.DataFrame(dict(text=TEXT_DATA))
        applier = PandasLFApplier([first_is_name, has_verb])
        L = applier.apply(df, progress_bar=False)
        np.testing.assert_equal(L, L_TEXT_EXPECTED)
        self.assertEqual(len(spacy._cache), 2)


class TestDaskApplier(unittest.TestCase):
    def test_lf_applier_dask(self) -> None:
        df = pd.DataFrame(dict(num=DATA))
        df = dd.from_pandas(df, npartitions=2)
        applier = DaskLFApplier([f, g])
        L = applier.apply(df)
        np.testing.assert_equal(L, L_EXPECTED)

    def test_lf_applier_dask_fault(self) -> None:
        df = pd.DataFrame(dict(num=DATA))
        df = dd.from_pandas(df, npartitions=2)
        applier = DaskLFApplier([f, f_bad])
        with self.assertRaises(Exception):
            applier.apply(df)
        L = applier.apply(df, fault_tolerant=True)
        np.testing.assert_equal(L, L_EXPECTED_BAD)

    def test_lf_applier_dask_preprocessor(self) -> None:
        df = pd.DataFrame(dict(num=DATA))
        df = dd.from_pandas(df, npartitions=2)
        applier = DaskLFApplier([f, fp])
        L = applier.apply(df)
        np.testing.assert_equal(L, L_PREPROCESS_EXPECTED)

    def test_lf_applier_pandas_preprocessor_memoized(self) -> None:
        @preprocessor(memoize=True)
        def square_memoize(x: DataPoint) -> DataPoint:
            x.num_squared = x.num**2
            return x

        @labeling_function(pre=[square_memoize])
        def fp_memoized(x: DataPoint) -> int:
            return 0 if x.num_squared > 42 else -1

        df = pd.DataFrame(dict(num=DATA))
        df = dd.from_pandas(df, npartitions=2)
        applier = DaskLFApplier([f, fp_memoized])
        L = applier.apply(df)
        np.testing.assert_equal(L, L_PREPROCESS_EXPECTED)

    @pytest.mark.complex
    def test_lf_applier_dask_spacy_preprocessor(self) -> None:
        spacy = SpacyPreprocessor(text_field="text", doc_field="doc")

        @labeling_function(pre=[spacy])
        def first_is_name(x: DataPoint) -> int:
            return 0 if x.doc[0].pos_ == "PROPN" else -1

        @labeling_function(pre=[spacy])
        def has_verb(x: DataPoint) -> int:
            return 0 if sum(t.pos_ == "VERB" for t in x.doc) > 0 else -1

        df = pd.DataFrame(dict(text=TEXT_DATA))
        df = dd.from_pandas(df, npartitions=2)
        applier = DaskLFApplier([first_is_name, has_verb])
        L = applier.apply(df)
        np.testing.assert_equal(L, L_TEXT_EXPECTED)

    @pytest.mark.complex
    def test_lf_applier_pandas_spacy_preprocessor_memoized(self) -> None:
        spacy = SpacyPreprocessor(text_field="text", doc_field="doc")
        spacy.memoize = True

        @labeling_function(pre=[spacy])
        def first_is_name(x: DataPoint) -> int:
            return 0 if x.doc[0].pos_ == "PROPN" else -1

        @labeling_function(pre=[spacy])
        def has_verb(x: DataPoint) -> int:
            return 0 if sum(t.pos_ == "VERB" for t in x.doc) > 0 else -1

        df = pd.DataFrame(dict(text=TEXT_DATA))
        df = dd.from_pandas(df, npartitions=2)
        applier = DaskLFApplier([first_is_name, has_verb])
        L = applier.apply(df)
        np.testing.assert_equal(L, L_TEXT_EXPECTED)

    def test_lf_applier_pandas_parallel(self) -> None:
        df = pd.DataFrame(dict(num=DATA))
        applier = PandasParallelLFApplier([f, g])
        L = applier.apply(df, n_parallel=2)
        np.testing.assert_equal(L, L_EXPECTED)

    def test_lf_applier_pandas_parallel_raises(self) -> None:
        df = pd.DataFrame(dict(num=DATA))
        applier = PandasParallelLFApplier([f, g])
        with self.assertRaises(ValueError):
            applier.apply(df, n_parallel=1)



import unittest

import numpy as np

from snorkel.utils import (
    filter_labels,
    preds_to_probs,
    probs_to_preds,
    to_int_label_array,
)

PROBS = np.array([[0.1, 0.9], [0.7, 0.3]])
PREDS = np.array([1, 0])
PREDS_ROUND = np.array([[0, 1], [1, 0]])


class UtilsTest(unittest.TestCase):
    def test_to_int_label_array(self):
        X = np.array([[1], [0], [2.0]])
        Y_expected = np.array([1, 0, 2])
        Y = to_int_label_array(X, flatten_vector=True)
        np.testing.assert_array_equal(Y, Y_expected)

        Y = to_int_label_array(np.array([[1]]), flatten_vector=True)
        Y_expected = np.array([1])
        np.testing.assert_array_equal(Y, Y_expected)

        Y = to_int_label_array(X, flatten_vector=False)
        Y_expected = np.array([[1], [0], [2]])
        np.testing.assert_array_equal(Y, Y_expected)

        X = np.array([[1], [0], [2.1]])
        with self.assertRaisesRegex(ValueError, "non-integer value"):
            to_int_label_array(X)

        X = np.array([[1, 0], [0, 1]])
        with self.assertRaisesRegex(ValueError, "1d np.array"):
            to_int_label_array(X, flatten_vector=True)

    def test_preds_to_probs(self):
        np.testing.assert_array_equal(preds_to_probs(PREDS, 2), PREDS_ROUND)

    def test_probs_to_preds(self):
        np.testing.assert_array_equal(probs_to_preds(PROBS), PREDS)

        # abtains with ties
        probs = np.array([[0.33, 0.33, 0.33]])
        preds = probs_to_preds(probs, tie_break_policy="abstain")
        true_preds = np.array([-1])
        np.testing.assert_array_equal(preds, true_preds)

        # true random with ties
        probs = np.array([[0.33, 0.33, 0.33]])
        random_preds = []
        for seed in range(10):
            preds = probs_to_preds(probs, tie_break_policy="true-random")
            random_preds.append(preds[0])

        # check predicted labels within range
        self.assertLessEqual(max(random_preds), 2)
        self.assertGreaterEqual(min(random_preds), 0)

        # deterministic random with ties
        probs = np.array(
            [[0.33, 0.33, 0.33], [0.0, 0.5, 0.5], [0.33, 0.33, 0.33], [0.5, 0.5, 0]]
        )
        random_preds = []
        for _ in range(10):
            preds = probs_to_preds(probs, tie_break_policy="random")
            random_preds.append(preds)

        # check labels are same across seeds
        for i in range(len(random_preds) - 1):
            np.testing.assert_array_equal(random_preds[i], random_preds[i + 1])

        # check predicted labels within range (only one instance since should all be same)
        self.assertLessEqual(max(random_preds[0]), 2)
        self.assertGreaterEqual(min(random_preds[0]), 0)

        # check invalid policy
        with self.assertRaisesRegex(ValueError, "policy not recognized"):
            preds = probs_to_preds(probs, tie_break_policy="negative")

        # check invalid input
        with self.assertRaisesRegex(ValueError, "probs must have probabilities"):
            preds = probs_to_preds(np.array([[0.33], [0.33]]))

    def test_filter_labels(self):
        golds = np.array([-1, 0, 0, 1, 1])
        preds = np.array([0, 0, 1, 1, -1])
        filtered = filter_labels(
            label_dict={"golds": golds, "preds": preds},
            filter_dict={"golds": [-1], "preds": [-1]},
        )
        np.testing.assert_array_equal(filtered["golds"], np.array([0, 0, 1]))
        np.testing.assert_array_equal(filtered["preds"], np.array([0, 1, 1]))

    def test_filter_labels_probs(self):
        golds = np.array([-1, 0, 0, 1, 1])
        preds = np.array([0, 0, 1, 1, -1])
        probs = np.array([[0.8, 0.2], [0.8, 0.2], [0.2, 0.8], [0.2, 0.8], [0.5, 0.5]])
        filtered = filter_labels(
            label_dict={"golds": golds, "preds": preds, "probs": probs},
            filter_dict={"golds": [-1], "preds": [-1]},
        )
        np.testing.assert_array_equal(filtered["golds"], np.array([0, 0, 1]))
        np.testing.assert_array_equal(filtered["preds"], np.array([0, 1, 1]))


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import List

import numpy as np
import pandas as pd
import pytest
from pyspark import SparkContext
from pyspark.sql import Row, SQLContext

from snorkel.labeling import labeling_function
from snorkel.labeling.apply.spark import SparkLFApplier
from snorkel.preprocess import preprocessor
from snorkel.types import DataPoint


@preprocessor()
def square(x: Row) -> Row:
    return Row(num=x.num, num_squared=x.num**2)


@labeling_function()
def f(x: DataPoint) -> int:
    return 0 if x.num > 42 else -1


@labeling_function(pre=[square])
def fp(x: DataPoint) -> int:
    return 0 if x.num_squared > 42 else -1


@labeling_function(resources=dict(db=[3, 6, 9]))
def g(x: DataPoint, db: List[int]) -> int:
    return 0 if x.num in db else -1


@labeling_function()
def f_bad(x: DataPoint) -> int:
    return 0 if x.mum > 42 else -1


DATA = [3, 43, 12, 9, 3]
L_EXPECTED = np.array([[-1, 0], [0, -1], [-1, -1], [-1, 0], [-1, 0]])
L_EXPECTED_BAD = np.array([[-1, -1], [0, -1], [-1, -1], [-1, -1], [-1, -1]])
L_PREPROCESS_EXPECTED = np.array([[-1, -1], [0, 0], [-1, 0], [-1, 0], [-1, -1]])

TEXT_DATA = ["Jane", "Jane plays soccer.", "Jane plays soccer."]
L_TEXT_EXPECTED = np.array([[0, -1], [0, 0], [0, 0]])


class TestSparkApplier(unittest.TestCase):
    @pytest.mark.complex
    @pytest.mark.spark
    def test_lf_applier_spark(self) -> None:
        sc = SparkContext.getOrCreate()
        sql = SQLContext(sc)
        df = pd.DataFrame(dict(num=DATA))
        rdd = sql.createDataFrame(df).rdd
        applier = SparkLFApplier([f, g])
        L = applier.apply(rdd)
        np.testing.assert_equal(L, L_EXPECTED)

    @pytest.mark.complex
    @pytest.mark.spark
    def test_lf_applier_spark_fault(self) -> None:
        sc = SparkContext.getOrCreate()
        sql = SQLContext(sc)
        df = pd.DataFrame(dict(num=DATA))
        rdd = sql.createDataFrame(df).rdd
        applier = SparkLFApplier([f, f_bad])
        with self.assertRaises(Exception):
            applier.apply(rdd)
        L = applier.apply(rdd, fault_tolerant=True)
        np.testing.assert_equal(L, L_EXPECTED_BAD)

    @pytest.mark.complex
    @pytest.mark.spark
    def test_lf_applier_spark_preprocessor(self) -> None:
        sc = SparkContext.getOrCreate()
        sql = SQLContext(sc)
        df = pd.DataFrame(dict(num=DATA))
        rdd = sql.createDataFrame(df).rdd
        applier = SparkLFApplier([f, fp])
        L = applier.apply(rdd)
        np.testing.assert_equal(L, L_PREPROCESS_EXPECTED)

    @pytest.mark.complex
    @pytest.mark.spark
    def test_lf_applier_spark_preprocessor_memoized(self) -> None:
        sc = SparkContext.getOrCreate()
        sql = SQLContext(sc)

        @preprocessor(memoize=True)
        def square_memoize(x: DataPoint) -> DataPoint:
            return Row(num=x.num, num_squared=x.num**2)

        @labeling_function(pre=[square_memoize])
        def fp_memoized(x: DataPoint) -> int:
            return 0 if x.num_squared > 42 else -1

        df = pd.DataFrame(dict(num=DATA))
        rdd = sql.createDataFrame(df).rdd
        applier = SparkLFApplier([f, fp_memoized])
        L = applier.apply(rdd)
        np.testing.assert_equal(L, L_PREPROCESS_EXPECTED)



import unittest
from types import SimpleNamespace
from typing import List

from snorkel.preprocess import preprocessor
from snorkel.slicing import SFApplier, slicing_function
from snorkel.types import DataPoint


@preprocessor()
def square(x: DataPoint) -> DataPoint:
    x.num_squared = x.num**2
    return x


class SquareHitTracker:
    def __init__(self):
        self.n_hits = 0

    def __call__(self, x: float) -> float:
        self.n_hits += 1
        return x**2


@slicing_function()
def f(x: DataPoint) -> int:
    return x.num > 42


@slicing_function(pre=[square])
def fp(x: DataPoint) -> int:
    return x.num_squared > 42


@slicing_function(resources=dict(db=[3, 6, 9]))
def g(x: DataPoint, db: List[int]) -> int:
    return x.num in db


DATA = [3, 43, 12, 9, 3]
S_EXPECTED = {"f": [0, 1, 0, 0, 0], "g": [1, 0, 0, 1, 1]}
S_PREPROCESS_EXPECTED = {"f": [0, 1, 0, 0, 0], "fp": [0, 1, 1, 1, 0]}


class TestSFApplier(unittest.TestCase):
    def test_sf_applier(self) -> None:
        data_points = [SimpleNamespace(num=num) for num in DATA]
        applier = SFApplier([f, g])
        S = applier.apply(data_points, progress_bar=False)
        self.assertEqual(S["f"].tolist(), S_EXPECTED["f"])
        self.assertEqual(S["g"].tolist(), S_EXPECTED["g"])
        S = applier.apply(data_points, progress_bar=True)
        self.assertEqual(S["f"].tolist(), S_EXPECTED["f"])
        self.assertEqual(S["g"].tolist(), S_EXPECTED["g"])

    def test_sf_applier_preprocessor(self) -> None:
        data_points = [SimpleNamespace(num=num) for num in DATA]
        applier = SFApplier([f, fp])
        S = applier.apply(data_points, progress_bar=False)
        self.assertEqual(S["f"].tolist(), S_PREPROCESS_EXPECTED["f"])
        self.assertEqual(S["fp"].tolist(), S_PREPROCESS_EXPECTED["fp"])



import unittest

import torch

from snorkel.classification.utils import (
    collect_flow_outputs_by_suffix,
    list_to_tensor,
    pad_batch,
)


class UtilsTest(unittest.TestCase):
    def test_pad_batch(self):
        batch = [torch.Tensor([1, 2]), torch.Tensor([3]), torch.Tensor([4, 5, 6])]
        padded_batch, mask_batch = pad_batch(batch)

        self.assertTrue(
            torch.equal(padded_batch, torch.Tensor([[1, 2, 0], [3, 0, 0], [4, 5, 6]]))
        )
        self.assertTrue(
            torch.equal(mask_batch, torch.Tensor([[0, 0, 1], [0, 1, 1], [0, 0, 0]]))
        )

        padded_batch, mask_batch = pad_batch(batch, max_len=2)

        self.assertTrue(
            torch.equal(padded_batch, torch.Tensor([[1, 2], [3, 0], [4, 5]]))
        )
        self.assertTrue(torch.equal(mask_batch, torch.Tensor([[0, 0], [0, 1], [0, 0]])))

        padded_batch, mask_batch = pad_batch(batch, pad_value=-1)

        self.assertTrue(
            torch.equal(
                padded_batch, torch.Tensor([[1, 2, -1], [3, -1, -1], [4, 5, 6]])
            )
        )
        self.assertTrue(
            torch.equal(mask_batch, torch.Tensor([[0, 0, 1], [0, 1, 1], [0, 0, 0]]))
        )

        padded_batch, mask_batch = pad_batch(batch, left_padded=True)

        self.assertTrue(
            torch.equal(padded_batch, torch.Tensor([[0, 1, 2], [0, 0, 3], [4, 5, 6]]))
        )
        self.assertTrue(
            torch.equal(mask_batch, torch.Tensor([[1, 0, 0], [1, 1, 0], [0, 0, 0]]))
        )

        padded_batch, mask_batch = pad_batch(batch, max_len=2, left_padded=True)

        self.assertTrue(
            torch.equal(padded_batch, torch.Tensor([[1, 2], [0, 3], [5, 6]]))
        )
        self.assertTrue(torch.equal(mask_batch, torch.Tensor([[0, 0], [1, 0], [0, 0]])))

    def test_list_to_tensor(self):
        # list of 1-D tensor with the different length
        batch = [torch.Tensor([1, 2]), torch.Tensor([3]), torch.Tensor([4, 5, 6])]

        padded_batch = list_to_tensor(batch)

        self.assertTrue(
            torch.equal(padded_batch, torch.Tensor([[1, 2, 0], [3, 0, 0], [4, 5, 6]]))
        )

        # list of 1-D tensor with the same length
        batch = [
            torch.Tensor([1, 2, 3]),
            torch.Tensor([4, 5, 6]),
            torch.Tensor([7, 8, 9]),
        ]

        padded_batch = list_to_tensor(batch)

        self.assertTrue(
            torch.equal(padded_batch, torch.Tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))
        )

        # list of 2-D tensor with the same size
        batch = [
            torch.Tensor([[1, 2, 3], [1, 2, 3]]),
            torch.Tensor([[4, 5, 6], [4, 5, 6]]),
            torch.Tensor([[7, 8, 9], [7, 8, 9]]),
        ]

        padded_batch = list_to_tensor(batch)

        self.assertTrue(
            torch.equal(
                padded_batch,
                torch.Tensor(
                    [
                        [[1, 2, 3], [1, 2, 3]],
                        [[4, 5, 6], [4, 5, 6]],
                        [[7, 8, 9], [7, 8, 9]],
                    ]
                ),
            )
        )

        # list of tensor with the different size
        batch = [
            torch.Tensor([[1, 2], [2, 3]]),
            torch.Tensor([4, 5, 6]),
            torch.Tensor([7, 8, 9, 0]),
        ]

        padded_batch = list_to_tensor(batch)

        self.assertTrue(
            torch.equal(
                padded_batch, torch.Tensor([[1, 2, 2, 3], [4, 5, 6, 0], [7, 8, 9, 0]])
            )
        )

    def test_collect_flow_outputs_by_suffix(self):
        flow_dict = {
            "a_pred_head": torch.Tensor([1]),
            "b_pred_head": torch.Tensor([2]),
            "c_pred": torch.Tensor([3]),
        }
        outputs = collect_flow_outputs_by_suffix(flow_dict, "_head")
        self.assertIn(torch.Tensor([1]), outputs)
        self.assertIn(torch.Tensor([2]), outputs)

    if __name__ == "__main__":
        unittest.main()



import unittest

import numpy as np
import torch
import torch.nn.functional as F

from snorkel.classification import cross_entropy_with_probs
from snorkel.utils import preds_to_probs


class SoftCrossEntropyTest(unittest.TestCase):
    def test_sce_equals_ce(self):
        DECIMAL = 6
        # Does soft ce loss match classic ce loss when labels are one-hot?
        Y_golds = torch.LongTensor([0, 1, 2])
        Y_golds_probs = torch.Tensor(preds_to_probs(Y_golds.numpy(), num_classes=4))

        Y_probs = torch.rand_like(Y_golds_probs)
        Y_probs = Y_probs / Y_probs.sum(dim=1).reshape(-1, 1)

        ce_loss = F.cross_entropy(Y_probs, Y_golds, reduction="none")
        ces_loss = cross_entropy_with_probs(Y_probs, Y_golds_probs, reduction="none")
        np.testing.assert_almost_equal(
            ce_loss.numpy(), ces_loss.numpy(), decimal=DECIMAL
        )

        ce_loss = F.cross_entropy(Y_probs, Y_golds, reduction="sum")
        ces_loss = cross_entropy_with_probs(Y_probs, Y_golds_probs, reduction="sum")
        np.testing.assert_almost_equal(
            ce_loss.numpy(), ces_loss.numpy(), decimal=DECIMAL
        )

        ce_loss = F.cross_entropy(Y_probs, Y_golds, reduction="mean")
        ces_loss = cross_entropy_with_probs(Y_probs, Y_golds_probs, reduction="mean")
        np.testing.assert_almost_equal(
            ce_loss.numpy(), ces_loss.numpy(), decimal=DECIMAL
        )

    def test_perfect_predictions(self):
        # Does soft ce loss achieve approx. 0 loss with perfect predictions?
        Y_golds = torch.LongTensor([0, 1, 2])
        Y_golds_probs = torch.Tensor(preds_to_probs(Y_golds.numpy(), num_classes=4))

        Y_probs = Y_golds_probs.clone()
        Y_probs[Y_probs == 1] = 100
        Y_probs[Y_probs == 0] = -100

        ces_loss = cross_entropy_with_probs(Y_probs, Y_golds_probs)
        np.testing.assert_equal(ces_loss.numpy(), 0)

    def test_lower_loss(self):
        # Is loss lower when it should be?
        Y_golds_probs = torch.tensor([[0.1, 0.9], [0.5, 0.5]])
        Y_probs1 = torch.tensor([[0.1, 0.3], [1.0, 0.0]])
        Y_probs2 = torch.tensor([[0.1, 0.2], [1.0, 0.0]])
        ces_loss1 = cross_entropy_with_probs(Y_probs1, Y_golds_probs)
        ces_loss2 = cross_entropy_with_probs(Y_probs2, Y_golds_probs)
        self.assertLess(ces_loss1, ces_loss2)

    def test_equal_loss(self):
        # Is loss equal when it should be?
        Y_golds_probs = torch.tensor([[0.1, 0.9], [0.5, 0.5]])
        Y_probs1 = torch.tensor([[0.1, 0.3], [1.0, 0.0]])
        Y_probs2 = torch.tensor([[0.1, 0.3], [0.0, 1.0]])
        ces_loss1 = cross_entropy_with_probs(Y_probs1, Y_golds_probs)
        ces_loss2 = cross_entropy_with_probs(Y_probs2, Y_golds_probs)
        self.assertEqual(ces_loss1, ces_loss2)

    def test_invalid_reduction(self):
        Y_golds = torch.LongTensor([0, 1, 2])
        Y_golds_probs = torch.Tensor(preds_to_probs(Y_golds.numpy(), num_classes=4))

        Y_probs = torch.rand_like(Y_golds_probs)
        Y_probs = Y_probs / Y_probs.sum(dim=1).reshape(-1, 1)

        with self.assertRaisesRegex(ValueError, "Keyword 'reduction' must be"):
            cross_entropy_with_probs(Y_probs, Y_golds_probs, reduction="bad")

    def test_loss_weights(self):
        FACTOR = 10

        # Do class weights work as expected?
        Y_golds = torch.LongTensor([0, 0, 1])
        Y_golds_probs = torch.Tensor(preds_to_probs(Y_golds.numpy(), num_classes=3))
        # Predict [1, 1, 1]
        Y_probs = torch.tensor(
            [[-100.0, 100.0, -100.0], [-100.0, 100.0, -100.0], [-100.0, 100.0, -100.0]]
        )

        ces_loss0 = cross_entropy_with_probs(Y_probs, Y_golds_probs).numpy()
        weight1 = torch.FloatTensor([1, 1, 1])
        ces_loss1 = cross_entropy_with_probs(
            Y_probs, Y_golds_probs, weight=weight1
        ).numpy()
        # Do weights of 1 match no weights at all?
        self.assertEqual(ces_loss0, ces_loss1)

        weight2 = torch.FloatTensor([1, 2, 1])
        ces_loss2 = cross_entropy_with_probs(
            Y_probs, Y_golds_probs, weight=weight2
        ).numpy()
        weight3 = weight2 * FACTOR
        ces_loss3 = cross_entropy_with_probs(
            Y_probs, Y_golds_probs, weight=weight3
        ).numpy()
        # If weights are X times larger, is loss X times larger?
        self.assertAlmostEqual(ces_loss2 * FACTOR, ces_loss3, places=3)

        # Note that PyTorch's cross-entropy loss has the unusual behavior that weights
        # behave differently when losses are averaged inside vs. outside the function.
        # See https://github.com/pytorch/pytorch/issues/8062 for details.
        ce_loss3 = (
            F.cross_entropy(Y_probs, Y_golds, weight=weight3, reduction="none")
            .mean()
            .numpy()
        )
        # Do hard and soft ce loss still match when we use class weights?
        self.assertAlmostEqual(ce_loss3, ces_loss3, places=3)


if __name__ == "__main__":
    unittest.main()



import unittest
from types import SimpleNamespace
from typing import List

import pandas as pd

from snorkel.augmentation import (
    ApplyOnePolicy,
    PandasTFApplier,
    RandomPolicy,
    TFApplier,
    transformation_function,
)
from snorkel.types import DataPoint


@transformation_function()
def square(x: DataPoint) -> DataPoint:
    x.num = x.num**2
    return x


@transformation_function()
def square_returns_none(x: DataPoint) -> DataPoint:
    if x.num == 2:
        return None
    x.num = x.num**2
    return x


@transformation_function()
def modify_in_place(x: DataPoint) -> DataPoint:
    x.d["my_key"] = 0
    return x


DATA = [1, 2, 3]
STR_DATA = ["x", "y", "z"]
DATA_IN_PLACE_EXPECTED = [(1 + i // 3) if i % 3 == 0 else 0 for i in range(9)]


def make_df(values: list, index: list, key: str = "num") -> pd.DataFrame:
    return pd.DataFrame({key: values}, index=index)


# NB: reconstruct each time to avoid inplace updates
def get_data_dict(data: List[int] = DATA):
    return [dict(my_key=num) for num in data]


class TestTFApplier(unittest.TestCase):
    def _get_x_namespace(self, data: List[int] = DATA) -> List[SimpleNamespace]:
        return [SimpleNamespace(num=num) for num in data]

    def _get_x_namespace_dict(self, data: List[int] = DATA) -> List[SimpleNamespace]:
        return [SimpleNamespace(d=d) for d in get_data_dict(data)]

    def test_tf_applier(self) -> None:
        data = self._get_x_namespace()
        policy = RandomPolicy(
            1, sequence_length=2, n_per_original=1, keep_original=False
        )
        applier = TFApplier([square], policy)
        data_augmented = applier.apply(data, progress_bar=False)
        self.assertEqual(data_augmented, self._get_x_namespace([1, 16, 81]))
        self.assertEqual(data, self._get_x_namespace())

        data_augmented = applier.apply(data, progress_bar=True)
        self.assertEqual(data_augmented, self._get_x_namespace([1, 16, 81]))
        self.assertEqual(data, self._get_x_namespace())

    def test_tf_applier_keep_original(self) -> None:
        data = self._get_x_namespace()
        policy = RandomPolicy(
            1, sequence_length=2, n_per_original=2, keep_original=True
        )
        applier = TFApplier([square], policy)
        data_augmented = applier.apply(data, progress_bar=False)
        vals = [1, 1, 1, 2, 16, 16, 3, 81, 81]
        self.assertEqual(data_augmented, self._get_x_namespace(vals))
        self.assertEqual(data, self._get_x_namespace())

    def test_tf_applier_returns_none(self) -> None:
        data = self._get_x_namespace()
        policy = RandomPolicy(
            1, sequence_length=2, n_per_original=2, keep_original=True
        )
        applier = TFApplier([square_returns_none], policy)
        data_augmented = applier.apply(data, progress_bar=False)
        vals = [1, 1, 1, 2, 3, 81, 81]
        self.assertEqual(data_augmented, self._get_x_namespace(vals))
        self.assertEqual(data, self._get_x_namespace())

    def test_tf_applier_keep_original_modify_in_place(self) -> None:
        data = self._get_x_namespace_dict()
        policy = ApplyOnePolicy(n_per_original=2, keep_original=True)
        applier = TFApplier([modify_in_place], policy)
        data_augmented = applier.apply(data, progress_bar=False)
        self.assertEqual(
            data_augmented, self._get_x_namespace_dict(DATA_IN_PLACE_EXPECTED)
        )
        self.assertEqual(data, self._get_x_namespace_dict())

    def test_tf_applier_generator(self) -> None:
        data = self._get_x_namespace()
        policy = RandomPolicy(
            1, sequence_length=2, n_per_original=2, keep_original=False
        )
        applier = TFApplier([square], policy)
        batches_expected = [[1, 1, 16, 16], [81, 81]]
        gen = applier.apply_generator(data, batch_size=2)
        for batch, batch_expected in zip(gen, batches_expected):
            self.assertEqual(batch, self._get_x_namespace(batch_expected))
        self.assertEqual(data, self._get_x_namespace())

    def test_tf_applier_keep_original_generator(self) -> None:
        data = self._get_x_namespace()
        policy = RandomPolicy(
            1, sequence_length=2, n_per_original=2, keep_original=True
        )
        applier = TFApplier([square], policy)
        batches_expected = [[1, 1, 1, 2, 16, 16], [3, 81, 81]]
        gen = applier.apply_generator(data, batch_size=2)
        for batch, batch_expected in zip(gen, batches_expected):
            self.assertEqual(batch, self._get_x_namespace(batch_expected))
        self.assertEqual(data, self._get_x_namespace())

    def test_tf_applier_returns_none_generator(self) -> None:
        data = self._get_x_namespace()
        policy = RandomPolicy(
            1, sequence_length=2, n_per_original=2, keep_original=True
        )
        applier = TFApplier([square_returns_none], policy)
        batches_expected = [[1, 1, 1, 2], [3, 81, 81]]
        gen = applier.apply_generator(data, batch_size=2)
        for batch, batch_expected in zip(gen, batches_expected):
            self.assertEqual(batch, self._get_x_namespace(batch_expected))
        self.assertEqual(data, self._get_x_namespace())

    def test_tf_applier_keep_original_modify_in_place_generator(self) -> None:
        data = self._get_x_namespace_dict()
        policy = ApplyOnePolicy(n_per_original=2, keep_original=True)
        applier = TFApplier([modify_in_place], policy)
        batches_expected = [DATA_IN_PLACE_EXPECTED[:6], DATA_IN_PLACE_EXPECTED[6:]]
        gen = applier.apply_generator(data, batch_size=2)
        for batch, batch_expected in zip(gen, batches_expected):
            self.assertEqual(batch, self._get_x_namespace_dict(batch_expected))
        self.assertEqual(data, self._get_x_namespace_dict())


class TestPandasTFApplier(unittest.TestCase):
    def _get_x_df(self):
        return pd.DataFrame(dict(num=DATA))

    def _get_x_df_with_str(self):
        return pd.DataFrame(dict(num=DATA, strs=STR_DATA))

    def _get_x_df_dict(self):
        return pd.DataFrame(dict(d=get_data_dict()))

    def test_tf_applier_pandas(self):
        df = self._get_x_df_with_str()
        policy = RandomPolicy(
            1, sequence_length=2, n_per_original=1, keep_original=False
        )
        applier = PandasTFApplier([square], policy)
        df_augmented = applier.apply(df, progress_bar=False)
        df_expected = pd.DataFrame(
            dict(num=[1, 16, 81], strs=STR_DATA), index=[0, 1, 2]
        )
        self.assertEqual(df_augmented.num.dtype, "int64")
        pd.testing.assert_frame_equal(df_augmented, df_expected)
        pd.testing.assert_frame_equal(df, self._get_x_df_with_str())

        df_augmented = applier.apply(df, progress_bar=True)
        df_expected = pd.DataFrame(
            dict(num=[1, 16, 81], strs=STR_DATA), index=[0, 1, 2]
        )
        pd.testing.assert_frame_equal(df_augmented, df_expected)
        pd.testing.assert_frame_equal(df, self._get_x_df_with_str())

    def test_tf_applier_pandas_keep_original(self):
        df = self._get_x_df()
        policy = RandomPolicy(
            1, sequence_length=2, n_per_original=2, keep_original=True
        )
        applier = PandasTFApplier([square], policy)
        df_augmented = applier.apply(df, progress_bar=False)
        df_expected = pd.DataFrame(
            dict(num=[1, 1, 1, 2, 16, 16, 3, 81, 81]), index=[0, 0, 0, 1, 1, 1, 2, 2, 2]
        )
        self.assertEqual(df_augmented.num.dtype, "int64")
        pd.testing.assert_frame_equal(df_augmented, df_expected)
        pd.testing.assert_frame_equal(df, self._get_x_df())

    def test_tf_applier_returns_none(self):
        df = self._get_x_df()
        policy = RandomPolicy(
            1, sequence_length=2, n_per_original=2, keep_original=True
        )
        applier = PandasTFApplier([square_returns_none], policy)
        df_augmented = applier.apply(df, progress_bar=False)
        df_expected = pd.DataFrame(
            dict(num=[1, 1, 1, 2, 3, 81, 81]), index=[0, 0, 0, 1, 2, 2, 2]
        )
        self.assertEqual(df_augmented.num.dtype, "int64")
        pd.testing.assert_frame_equal(df_augmented, df_expected)
        pd.testing.assert_frame_equal(df, self._get_x_df())

    def test_tf_applier_pandas_modify_in_place(self):
        df = self._get_x_df_dict()
        policy = ApplyOnePolicy(n_per_original=2, keep_original=True)
        applier = PandasTFApplier([modify_in_place], policy)
        df_augmented = applier.apply(df, progress_bar=False)
        idx = [0, 0, 0, 1, 1, 1, 2, 2, 2]
        df_expected = pd.DataFrame(
            dict(d=get_data_dict(DATA_IN_PLACE_EXPECTED)), index=idx
        )
        pd.testing.assert_frame_equal(df_augmented, df_expected)
        pd.testing.assert_frame_equal(df, self._get_x_df_dict())

    def test_tf_applier_pandas_generator(self):
        df = self._get_x_df_with_str()
        policy = RandomPolicy(
            1, sequence_length=2, n_per_original=2, keep_original=False
        )
        applier = PandasTFApplier([square], policy)
        gen = applier.apply_generator(df, batch_size=2)
        df_expected = [
            pd.DataFrame(
                {"num": [1, 1, 16, 16], "strs": ["x", "x", "y", "y"]},
                index=[0, 0, 1, 1],
            ),
            pd.DataFrame({"num": [81, 81], "strs": ["z", "z"]}, index=[2, 2]),
        ]
        for df_batch, df_batch_expected in zip(gen, df_expected):
            self.assertEqual(df_batch.num.dtype, "int64")
            pd.testing.assert_frame_equal(df_batch, df_batch_expected)
        pd.testing.assert_frame_equal(df, self._get_x_df_with_str())

    def test_tf_applier_pandas_keep_original_generator(self):
        df = self._get_x_df()
        policy = RandomPolicy(
            1, sequence_length=2, n_per_original=2, keep_original=True
        )
        applier = PandasTFApplier([square], policy)
        gen = applier.apply_generator(df, batch_size=2)
        df_expected = [
            make_df([1, 1, 1, 2, 16, 16], [0, 0, 0, 1, 1, 1]),
            make_df([3, 81, 81], [2, 2, 2]),
        ]
        for df_batch, df_batch_expected in zip(gen, df_expected):
            pd.testing.assert_frame_equal(df_batch, df_batch_expected)
        pd.testing.assert_frame_equal(df, self._get_x_df())

    def test_tf_applier_returns_none_generator(self):
        df = self._get_x_df()
        policy = RandomPolicy(
            1, sequence_length=2, n_per_original=2, keep_original=True
        )
        applier = PandasTFApplier([square_returns_none], policy)
        gen = applier.apply_generator(df, batch_size=2)
        df_expected = [
            make_df([1, 1, 1, 2], [0, 0, 0, 1]),
            make_df([3, 81, 81], [2, 2, 2]),
        ]
        for df_batch, df_batch_expected in zip(gen, df_expected):
            pd.testing.assert_frame_equal(df_batch, df_batch_expected)
        pd.testing.assert_frame_equal(df, self._get_x_df())

    def test_tf_applier_pandas_modify_in_place_generator(self):
        df = self._get_x_df_dict()
        policy = ApplyOnePolicy(n_per_original=2, keep_original=True)
        applier = PandasTFApplier([modify_in_place], policy)
        gen = applier.apply_generator(df, batch_size=2)
        idx = [0, 0, 0, 1, 1, 1, 2, 2, 2]
        df_expected = [
            make_df(get_data_dict(DATA_IN_PLACE_EXPECTED[:6]), idx[:6], key="d"),
            make_df(get_data_dict(DATA_IN_PLACE_EXPECTED[6:]), idx[6:], key="d"),
        ]
        for df_batch, df_batch_expected in zip(gen, df_expected):
            pd.testing.assert_frame_equal(df_batch, df_batch_expected)
        pd.testing.assert_frame_equal(df, self._get_x_df_dict())



import unittest

import pandas as pd

from snorkel.slicing import slicing_function
from snorkel.slicing.monitor import slice_dataframe

DATA = [5, 10, 19, 22, 25]


@slicing_function()
def sf(x):
    return x.num < 20


class PandasSlicerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.df = pd.DataFrame(dict(num=DATA))

    def test_slice(self):
        self.assertEqual(len(self.df), 5)

        # Should return a subset
        sliced_df = slice_dataframe(self.df, sf)
        self.assertEqual(len(sliced_df), 3)



from typing import Tuple

import numpy as np
from pyspark import RDD

from snorkel.types import DataPoint

from .core import BaseLFApplier, RowData, _FunctionCaller, apply_lfs_to_data_point


class SparkLFApplier(BaseLFApplier):
    r"""LF applier for a Spark RDD.

    Data points are stored as ``Row``\s in an RDD, and a Spark
    ``map`` job is submitted to execute the LFs. A common
    way to obtain an RDD is via a PySpark DataFrame. For an
    example usage with AWS EMR instructions, see
    ``test/labeling/apply/lf_applier_spark_test_script.py``.
    """

    def apply(self, data_points: RDD, fault_tolerant: bool = False) -> np.ndarray:
        """Label PySpark RDD of data points with LFs.

        Parameters
        ----------
        data_points
            PySpark RDD containing data points to be labeled by LFs
        fault_tolerant
            Output ``-1`` if LF execution fails?

        Returns
        -------
        np.ndarray
            Matrix of labels emitted by LFs
        """
        f_caller = _FunctionCaller(fault_tolerant)

        def map_fn(args: Tuple[DataPoint, int]) -> RowData:
            return apply_lfs_to_data_point(*args, lfs=self._lfs, f_caller=f_caller)

        labels = data_points.zipWithIndex().map(map_fn).collect()
        return self._numpy_from_row_data(labels)



import random
import unittest

import numpy as np
import torch

from snorkel.slicing import SliceCombinerModule


class SliceCombinerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        random.seed(123)
        np.random.seed(123)
        torch.manual_seed(123)

    def test_forward_shape(self):
        """Test that the reweight representation shape matches expected feature size."""

        # common case
        batch_size = 4
        h_dim = 20
        num_classes = 2

        outputs = {
            # Always 2-dim binary outputs
            "task_slice:base_ind_head": torch.FloatTensor(batch_size, 2).uniform_(0, 1),
            "task_slice:base_pred_transform": torch.FloatTensor(
                batch_size, h_dim
            ).uniform_(0, 1),
            "task_slice:base_pred_head": torch.FloatTensor(
                batch_size, num_classes
            ).uniform_(0, 1),
        }
        combiner_module = SliceCombinerModule()
        combined_rep = combiner_module(outputs)
        self.assertEqual(tuple(combined_rep.shape), (batch_size, h_dim))

        # edge case (some ones)
        batch_size = 1
        h_dim = 1
        num_classes = 2

        outputs = {
            # Always 2-dim binary outputs
            "task_slice:base_ind_head": torch.FloatTensor(batch_size, 2).uniform_(0, 1),
            "task_slice:base_pred_transform": torch.FloatTensor(
                batch_size, h_dim
            ).uniform_(0, 1),
            "task_slice:base_pred_head": torch.FloatTensor(
                batch_size, num_classes
            ).uniform_(0, 1),
        }
        combiner_module = SliceCombinerModule()
        combined_rep = combiner_module(outputs)
        self.assertEqual(tuple(combined_rep.shape), (batch_size, h_dim))

        # num_classes = 1
        batch_size = 1
        h_dim = 1
        num_classes = 1

        outputs = {
            # Always 2-dim binary outputs
            "task_slice:base_ind_head": torch.FloatTensor(batch_size, 2).uniform_(0, 1),
            "task_slice:base_pred_transform": torch.FloatTensor(
                batch_size, h_dim
            ).uniform_(0, 1),
            "task_slice:base_pred_head": torch.FloatTensor(
                batch_size, num_classes
            ).uniform_(0, 1),
        }
        combiner_module = SliceCombinerModule()
        with self.assertRaisesRegex(NotImplementedError, "requires output shape"):
            combined_rep = combiner_module(outputs)

    def test_average_reweighting(self):
        """Test average reweighting (equal weight across two slices)."""
        batch_size = 4
        h_dim = 20
        num_classes = 2

        outputs = {
            "task_slice:a_ind_head": torch.ones(batch_size, 2) * 10.0,
            "task_slice:a_pred_transform": torch.ones(batch_size, h_dim) * 4,
            "task_slice:a_pred_head": torch.ones(batch_size, num_classes) * 10.0,
            "task_slice:b_ind_head": torch.ones(batch_size, 2) * -10.0,
            "task_slice:b_pred_transform": torch.ones(batch_size, h_dim) * 2,
            "task_slice:b_pred_head": torch.ones(batch_size, num_classes) * -10.0,
        }
        combiner_module = SliceCombinerModule()
        combined_rep = combiner_module(outputs)
        self.assertTrue(torch.allclose(combined_rep, torch.ones(batch_size, h_dim) * 3))

    def test_average_reweighting_by_ind(self):
        """Test average reweighting by ind (zeros in pred head)."""
        batch_size = 4
        h_dim = 20
        num_classes = 2

        outputs = {
            "task_slice:a_ind_head": torch.ones(batch_size, 2) * 10.0,
            "task_slice:a_pred_transform": torch.ones(batch_size, h_dim) * 4,
            "task_slice:a_pred_head": torch.zeros(batch_size, num_classes),
            "task_slice:b_ind_head": torch.ones(batch_size, 2) * -10.0,
            "task_slice:b_pred_transform": torch.ones(batch_size, h_dim) * 2,
            "task_slice:b_pred_head": torch.zeros(batch_size, num_classes),
        }
        combiner_module = SliceCombinerModule()
        combined_rep = combiner_module(outputs)
        self.assertTrue(torch.allclose(combined_rep, torch.ones(batch_size, h_dim) * 3))

    def test_average_reweighting_by_pred_confidence(self):
        """Test average reweighting by pred confidence (zeros in ind head)."""
        batch_size = 4
        h_dim = 20
        num_classes = 2

        outputs = {
            "task_slice:a_ind_head": torch.zeros(batch_size, 2),
            "task_slice:a_pred_transform": torch.ones(batch_size, h_dim) * 4,
            "task_slice:a_pred_head": torch.ones(batch_size, num_classes) * 5,
            "task_slice:b_ind_head": torch.zeros(batch_size, 2),
            "task_slice:b_pred_transform": torch.ones(batch_size, h_dim) * 2,
            "task_slice:b_pred_head": torch.ones(batch_size, num_classes) * 5,
        }
        combiner_module = SliceCombinerModule()
        combined_rep = combiner_module(outputs)
        self.assertTrue(torch.all(combined_rep == torch.ones(batch_size, h_dim) * 3))

        # Changing sign of prediction shouldn't change confidence
        outputs = {
            "task_slice:a_ind_head": torch.zeros(batch_size, 2),
            "task_slice:a_pred_transform": torch.ones(batch_size, h_dim) * 4,
            "task_slice:a_pred_head": torch.ones(batch_size, num_classes) * -5,
            "task_slice:b_ind_head": torch.zeros(batch_size, 2),
            "task_slice:b_pred_transform": torch.ones(batch_size, h_dim) * 2,
            "task_slice:b_pred_head": torch.ones(batch_size, num_classes) * 5,
        }
        combiner_module = SliceCombinerModule()
        combined_rep = combiner_module(outputs)
        self.assertTrue(torch.allclose(combined_rep, torch.ones(batch_size, h_dim) * 3))

    def test_many_slices(self):
        """Test combiner on 100 synthetic generated slices."""

        batch_size = 4
        h_dim = 20
        num_classes = 2

        outputs = {}
        # Generate 100 slices, half voting in one direction (index i) and
        # half in another (index i%2), resulting in averaged weights
        for i in range(100):
            if i % 2 == 0:
                outputs[f"task_slice:{i}_ind_head"] = torch.ones(batch_size, 2) * 20.0
                outputs[f"task_slice:{i}_pred_transform"] = (
                    torch.ones(batch_size, h_dim) * 4
                )
                outputs[f"task_slice:{i}_pred_head"] = (
                    torch.ones(batch_size, num_classes) * 20.0
                )
            else:
                outputs[f"task_slice:{i}_ind_head"] = torch.ones(batch_size, 2) * -20.0
                outputs[f"task_slice:{i}_pred_transform"] = (
                    torch.ones(batch_size, h_dim) * 2
                )
                outputs[f"task_slice:{i}_pred_head"] = (
                    torch.ones(batch_size, num_classes) * -20.0
                )

        combiner_module = SliceCombinerModule()
        combined_rep = combiner_module(outputs)
        self.assertTrue(torch.allclose(combined_rep, torch.ones(batch_size, h_dim) * 3))

    def test_combiner_multiclass(self):
        """Test combiner in multiclass setting."""
        batch_size = 4
        h_dim = 20
        num_classes = 10

        # For each class, make one class (at random) the maximum score
        # For everything else, set to "logit" between [-5, 5]
        max_score_indexes_a = [
            random.randint(0, num_classes) for _ in range(batch_size)
        ]
        pred_outputs_a = torch.FloatTensor(batch_size, num_classes).uniform_(-5, 5)
        pred_outputs_a = pred_outputs_a.scatter_(
            # scatter max score (10.0) at indexes
            1,
            torch.tensor(max_score_indexes_a).unsqueeze(1),
            10.0,
        )

        max_score_indexes_b = [
            random.randint(0, num_classes) for _ in range(batch_size)
        ]
        pred_outputs_b = torch.FloatTensor(batch_size, num_classes).uniform_(-5, 5)
        pred_outputs_b = pred_outputs_b.scatter_(
            # scatter max score (-10.0) at indexes
            1,
            torch.tensor(max_score_indexes_b).unsqueeze(1),
            10.0,
        )

        outputs = {
            "task_slice:a_ind_head": torch.ones(batch_size, 2) * -10.0,
            "task_slice:a_pred_transform": torch.ones(batch_size, h_dim) * 4,
            "task_slice:a_pred_head": pred_outputs_a,
            "task_slice:b_ind_head": torch.ones(batch_size, 2) * 10.0,
            "task_slice:b_pred_transform": torch.ones(batch_size, h_dim) * 2,
            "task_slice:b_pred_head": pred_outputs_b,
        }

        # Ensure smoother temperature for multi-class
        combiner_module = SliceCombinerModule()
        with self.assertRaisesRegex(NotImplementedError, "more than 2 classes"):
            combiner_module(outputs)

    def test_temperature(self):
        """Test temperature parameter for attention weights."""
        batch_size = 4
        h_dim = 20
        num_classes = 2

        # Add noise to each set of inputs to attention weights
        epsilon = 1e-5
        outputs = {
            "task_slice:a_ind_head": torch.ones(batch_size, 2) * 10.0
            + torch.FloatTensor(batch_size, 2).normal_(0.0, epsilon),
            "task_slice:a_pred_transform": torch.ones(batch_size, h_dim) * 4
            + torch.FloatTensor(batch_size, h_dim).normal_(0.0, epsilon),
            "task_slice:a_pred_head": torch.ones(batch_size, num_classes) * 10.0
            + torch.FloatTensor(batch_size, num_classes).normal_(0.0, epsilon),
            "task_slice:b_ind_head": torch.ones(batch_size, 2) * -10.0
            + torch.FloatTensor(batch_size, 2).normal_(0.0, epsilon),
            "task_slice:b_pred_transform": torch.ones(batch_size, h_dim) * 2
            + torch.FloatTensor(batch_size, h_dim).normal_(0.0, epsilon),
            "task_slice:b_pred_head": torch.ones(batch_size, num_classes) * -10.0
            + torch.FloatTensor(batch_size, num_classes).normal_(0.0, epsilon),
        }

        # With larger temperature, attention is smoother and reweighted rep is closer
        # closer to true average, despite noise
        combiner_module = SliceCombinerModule(temperature=1e5)
        combined_rep = combiner_module(outputs)
        self.assertTrue(torch.allclose(combined_rep, torch.ones(batch_size, h_dim) * 3))

        # With (impractically) small temperature, attention peaks/biases towards in
        # some direction of noisy attention weights
        combiner_module = SliceCombinerModule(temperature=1e-15)
        combined_rep = combiner_module(outputs)

        # Check number of elements that match either of the original weights
        # Every example should either match 2 or 4
        isclose_four = torch.isclose(
            combined_rep, torch.ones(batch_size, h_dim) * 2, atol=1e-4
        )
        isclose_two = torch.isclose(
            combined_rep, torch.ones(batch_size, h_dim) * 4, atol=1e-4
        )
        num_matching_original = torch.sum(isclose_four) + torch.sum(isclose_two)
        self.assertEqual(num_matching_original, batch_size * h_dim)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import pandas as pd

from snorkel.labeling import LabelingFunction, LFAnalysis

L = [
    [-1, -1, 0, -1, -1, 0],
    [-1, -1, -1, 2, -1, -1],
    [2, -1, -1, -1, -1, 0],
    [1, -1, 2, -1, 0, 0],
    [-1, -1, -1, -1, -1, -1],
    [1, -1, 0, 2, 1, 0],
]

Y = [0, 1, 2, 0, 1, 2]

L_wo_abstain = [
    [3, 3, 3, 5, 4, 3],
    [3, 4, 5, 5, 4, 3],
    [5, 3, 4, 4, 5, 3],
    [4, 4, 5, 4, 3, 3],
    [3, 4, 3, 5, 4, 3],
    [5, 3, 3, 4, 4, 3],
]


def f(x):
    return -1


class TestAnalysis(unittest.TestCase):
    def setUp(self) -> None:
        self.lfa = LFAnalysis(np.array(L))
        self.lfa_wo_abstain = LFAnalysis(np.array(L_wo_abstain))
        self.Y = np.array(Y)

    def test_label_coverage(self) -> None:
        self.assertEqual(self.lfa.label_coverage(), 5 / 6)

    def test_label_overlap(self) -> None:
        self.assertEqual(self.lfa.label_overlap(), 4 / 6)

    def test_label_conflict(self) -> None:
        self.assertEqual(self.lfa.label_conflict(), 3 / 6)

    def test_lf_polarities(self) -> None:
        polarities = self.lfa.lf_polarities()
        self.assertEqual(polarities, [[1, 2], [], [0, 2], [2], [0, 1], [0]])

    def test_lf_coverages(self) -> None:
        coverages = self.lfa.lf_coverages()
        coverages_expected = [3 / 6, 0, 3 / 6, 2 / 6, 2 / 6, 4 / 6]
        np.testing.assert_array_almost_equal(coverages, np.array(coverages_expected))

    def test_lf_overlaps(self) -> None:
        overlaps = self.lfa.lf_overlaps(normalize_by_coverage=False)
        overlaps_expected = [3 / 6, 0, 3 / 6, 1 / 6, 2 / 6, 4 / 6]
        np.testing.assert_array_almost_equal(overlaps, np.array(overlaps_expected))

        overlaps = self.lfa.lf_overlaps(normalize_by_coverage=True)
        overlaps_expected = [1, 0, 1, 1 / 2, 1, 1]
        np.testing.assert_array_almost_equal(overlaps, np.array(overlaps_expected))

    def test_lf_conflicts(self) -> None:
        conflicts = self.lfa.lf_conflicts(normalize_by_overlaps=False)
        conflicts_expected = [3 / 6, 0, 2 / 6, 1 / 6, 2 / 6, 3 / 6]
        np.testing.assert_array_almost_equal(conflicts, np.array(conflicts_expected))

        conflicts = self.lfa.lf_conflicts(normalize_by_overlaps=True)
        conflicts_expected = [1, 0, 2 / 3, 1, 1, 3 / 4]
        np.testing.assert_array_almost_equal(conflicts, np.array(conflicts_expected))

    def test_lf_empirical_accuracies(self) -> None:
        accs = self.lfa.lf_empirical_accuracies(self.Y)
        accs_expected = [1 / 3, 0, 1 / 3, 1 / 2, 1 / 2, 2 / 4]
        np.testing.assert_array_almost_equal(accs, np.array(accs_expected))

    def test_lf_empirical_probs(self) -> None:
        P_emp = self.lfa.lf_empirical_probs(self.Y, 3)
        P = np.array(
            [
                [[1 / 2, 1, 0], [0, 0, 0], [1 / 2, 0, 1 / 2], [0, 0, 1 / 2]],
                [[1, 1, 1], [0, 0, 0], [0, 0, 0], [0, 0, 0]],
                [[0, 1, 1 / 2], [1 / 2, 0, 1 / 2], [0, 0, 0], [1 / 2, 0, 0]],
                [[1, 1 / 2, 1 / 2], [0, 0, 0], [0, 0, 0], [0, 1 / 2, 1 / 2]],
                [[1 / 2, 1, 1 / 2], [1 / 2, 0, 0], [0, 0, 1 / 2], [0, 0, 0]],
                [[0, 1, 0], [1, 0, 1], [0, 0, 0], [0, 0, 0]],
            ]
        )
        np.testing.assert_array_almost_equal(P, P_emp)

    def test_lf_summary(self) -> None:
        df = self.lfa.lf_summary(self.Y, est_weights=None)
        df_expected = pd.DataFrame(
            {
                "Polarity": [[1, 2], [], [0, 2], [2], [0, 1], [0]],
                "Coverage": [3 / 6, 0, 3 / 6, 2 / 6, 2 / 6, 4 / 6],
                "Overlaps": [3 / 6, 0, 3 / 6, 1 / 6, 2 / 6, 4 / 6],
                "Conflicts": [3 / 6, 0, 2 / 6, 1 / 6, 2 / 6, 3 / 6],
                "Correct": [1, 0, 1, 1, 1, 2],
                "Incorrect": [2, 0, 2, 1, 1, 2],
                "Emp. Acc.": [1 / 3, 0, 1 / 3, 1 / 2, 1 / 2, 2 / 4],
            }
        )
        pd.testing.assert_frame_equal(df.round(6), df_expected.round(6))

        df = self.lfa.lf_summary(Y=None, est_weights=None)
        df_expected = pd.DataFrame(
            {
                "Polarity": [[1, 2], [], [0, 2], [2], [0, 1], [0]],
                "Coverage": [3 / 6, 0, 3 / 6, 2 / 6, 2 / 6, 4 / 6],
                "Overlaps": [3 / 6, 0, 3 / 6, 1 / 6, 2 / 6, 4 / 6],
                "Conflicts": [3 / 6, 0, 2 / 6, 1 / 6, 2 / 6, 3 / 6],
            }
        )
        pd.testing.assert_frame_equal(df.round(6), df_expected.round(6))

        est_weights = [1, 0, 1, 1, 1, 0.5]
        names = list("abcdef")
        lfs = [LabelingFunction(s, f) for s in names]
        lfa = LFAnalysis(np.array(L), lfs)
        df = lfa.lf_summary(self.Y, est_weights=est_weights)
        df_expected = pd.DataFrame(
            {
                "j": [0, 1, 2, 3, 4, 5],
                "Polarity": [[1, 2], [], [0, 2], [2], [0, 1], [0]],
                "Coverage": [3 / 6, 0, 3 / 6, 2 / 6, 2 / 6, 4 / 6],
                "Overlaps": [3 / 6, 0, 3 / 6, 1 / 6, 2 / 6, 4 / 6],
                "Conflicts": [3 / 6, 0, 2 / 6, 1 / 6, 2 / 6, 3 / 6],
                "Correct": [1, 0, 1, 1, 1, 2],
                "Incorrect": [2, 0, 2, 1, 1, 2],
                "Emp. Acc.": [1 / 3, 0, 1 / 3, 1 / 2, 1 / 2, 2 / 4],
                "Learned Weight": [1, 0, 1, 1, 1, 0.5],
            }
        ).set_index(pd.Index(names))
        pd.testing.assert_frame_equal(df.round(6), df_expected.round(6))

    def test_wrong_number_of_lfs(self) -> None:
        with self.assertRaisesRegex(ValueError, "Number of LFs"):
            LFAnalysis(np.array(L), [LabelingFunction(s, f) for s in "ab"])

    def test_lf_summary_without_abstain(self) -> None:
        df = self.lfa_wo_abstain.lf_summary(self.Y + 4, est_weights=None)
        df_expected = pd.DataFrame(
            {
                "Polarity": [[3, 4, 5], [3, 4], [3, 4, 5], [4, 5], [3, 4, 5], [3]],
                "Coverage": [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                "Overlaps": [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                "Conflicts": [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                "Correct": [1, 1, 1, 3, 1, 0],
                "Incorrect": [5, 5, 5, 3, 5, 6],
                "Emp. Acc.": [1 / 6, 1 / 6, 1 / 6, 3 / 6, 1 / 6, 0],
            }
        )
        pd.testing.assert_frame_equal(df.round(6), df_expected.round(6))



import unittest

import numpy as np

from snorkel.labeling.model import MajorityClassVoter, MajorityLabelVoter, RandomVoter


class BaselineModelTest(unittest.TestCase):
    def test_random_vote(self):
        L = np.array([[0, 1, 0], [-1, 3, 2], [2, -1, -1], [0, 1, 1]])
        rand_voter = RandomVoter()
        Y_p = rand_voter.predict_proba(L)
        self.assertLessEqual(Y_p.max(), 1.0)
        self.assertGreaterEqual(Y_p.min(), 0.0)
        np.testing.assert_array_almost_equal(
            np.sum(Y_p, axis=1), np.ones(np.shape(L)[0])
        )

    def test_majority_class_vote(self):
        L = np.array([[0, 1, 0], [1, 1, 0], [1, 1, 0], [-1, -1, 1]])
        mc_voter = MajorityClassVoter()
        mc_voter.fit(balance=np.array([0.8, 0.2]))
        Y_p = mc_voter.predict_proba(L)

        Y_p_true = np.array([[1.0, 0.0], [1.0, 0.0], [1.0, 0.0], [1.0, 0.0]])
        np.testing.assert_array_almost_equal(Y_p, Y_p_true)

    def test_majority_label_vote(self):
        L = np.array([[0, 1, 0], [0, 1, 0], [1, 0, 0], [-1, -1, 1]])
        ml_voter = MajorityLabelVoter()
        Y_p = ml_voter.predict_proba(L)

        Y_p_true = np.array([[1.0, 0.0], [1.0, 0.0], [1.0, 0.0], [0.0, 1.0]])
        np.testing.assert_array_almost_equal(Y_p, Y_p_true)


if __name__ == "__main__":
    unittest.main()



import unittest
from types import SimpleNamespace
from typing import Any, Optional

import numpy as np
import pandas as pd
import spacy

from snorkel.map import Mapper, lambda_mapper
from snorkel.map.core import get_hashable
from snorkel.types import DataPoint, FieldMap


class SplitWordsMapper(Mapper):
    def __init__(
        self, name: str, text_field: str, lower_field: str, words_field: str
    ) -> None:
        super().__init__(
            name, dict(text=text_field), dict(lower=lower_field, words=words_field)
        )

    def run(self, text: str) -> FieldMap:  # type: ignore
        return dict(lower=text.lower(), words=text.split())


class SplitWordsMapperDefaultArgs(Mapper):
    def run(self, text: str) -> FieldMap:  # type: ignore
        return dict(lower=text.lower(), words=text.split())


class MapperReturnsNone(Mapper):
    def run(self, text: str) -> Optional[FieldMap]:  # type: ignore
        return None


class MapperWithArgs(Mapper):
    def run(self, text: str, *args: Any) -> Optional[FieldMap]:  # type: ignore
        return None


class MapperWithKwargs(Mapper):
    def run(self, text: str, **kwargs: Any) -> Optional[FieldMap]:  # type: ignore
        return None


class MapperWithPre(Mapper):
    def run(self, num_squared: float) -> Optional[FieldMap]:  # type: ignore
        return dict(double_num_squared=2 * num_squared)


class MapperWithPre2(Mapper):
    def run(self, double_num_squared: float) -> Optional[FieldMap]:  # type: ignore
        return dict(double_num_squared_plus_1=double_num_squared + 1)


class SquareHitTracker:
    def __init__(self):
        self.n_hits = 0

    def __call__(self, x: float) -> float:
        self.n_hits += 1
        return x**2


@lambda_mapper()
def square(x: DataPoint) -> DataPoint:
    x.num_squared = x.num**2
    return x


@lambda_mapper()
def modify_in_place(x: DataPoint) -> DataPoint:
    x.d["my_key"] = 0
    x.d_new = x.d
    return x


class TestMapperCore(unittest.TestCase):
    def _get_x(self, num=8, text="Henry has fun") -> SimpleNamespace:
        return SimpleNamespace(num=num, text=text)

    def _get_x_dict(self) -> SimpleNamespace:
        return SimpleNamespace(num=8, d=dict(my_key=1))

    def test_numeric_mapper(self) -> None:
        x_mapped = square(self._get_x())
        # NB: not using `self.assertIsNotNone` due to mypy
        # See https://github.com/python/mypy/issues/5088
        assert x_mapped is not None
        x_expected = SimpleNamespace(num=8, text="Henry has fun", num_squared=64)
        self.assertEqual(x_mapped, x_expected)

    def test_text_mapper(self) -> None:
        split_words = SplitWordsMapper(
            "split_words", "text", "text_lower", "text_words"
        )
        x_mapped = split_words(self._get_x())
        assert x_mapped is not None
        x_expected = SimpleNamespace(
            num=8,
            text="Henry has fun",
            text_lower="henry has fun",
            text_words=["Henry", "has", "fun"],
        )
        self.assertEqual(x_mapped, x_expected)

    def test_mapper_same_field(self) -> None:
        split_words = SplitWordsMapper("split_words", "text", "text", "text_words")
        x = self._get_x()
        x_mapped = split_words(x)
        self.assertEqual(x, self._get_x())
        assert x_mapped is not None
        x_expected = SimpleNamespace(
            num=8, text="henry has fun", text_words=["Henry", "has", "fun"]
        )
        self.assertEqual(x_mapped, x_expected)

    def test_mapper_default_args(self) -> None:
        split_words = SplitWordsMapperDefaultArgs("split_words")
        x_mapped = split_words(self._get_x())
        assert x_mapped is not None
        x_expected = SimpleNamespace(
            num=8,
            text="Henry has fun",
            lower="henry has fun",
            words=["Henry", "has", "fun"],
        )
        self.assertEqual(x_mapped, x_expected)

    def test_mapper_in_place(self) -> None:
        x = self._get_x_dict()
        x_mapped = modify_in_place(x)
        self.assertEqual(x, self._get_x_dict())
        assert x_mapped is not None
        x_expected = SimpleNamespace(num=8, d=dict(my_key=0), d_new=dict(my_key=0))
        self.assertEqual(x_mapped, x_expected)

    def test_mapper_returns_none(self) -> None:
        mapper = MapperReturnsNone("none_mapper")
        x_mapped = mapper(self._get_x())
        self.assertIsNone(x_mapped)

    def test_mapper_pre(self) -> None:
        mapper_no_pre = MapperWithPre("pre_mapper")
        x = self._get_x(3)
        with self.assertRaises(AttributeError):
            x_mapped = mapper_no_pre(x)

        mapper_pre = MapperWithPre("pre_mapper", pre=[square])
        x = self._get_x(3)
        x_mapped = mapper_pre(x)
        self.assertEqual(x, self._get_x(3))
        assert x_mapped is not None
        x_expected = SimpleNamespace(
            num=3, num_squared=9, double_num_squared=18, text="Henry has fun"
        )
        self.assertEqual(x_mapped, x_expected)

        mapper_pre_2 = MapperWithPre2("pre_mapper", pre=[mapper_pre])
        x = self._get_x(3)
        x_mapped = mapper_pre_2(x)
        self.assertEqual(x, self._get_x(3))
        assert x_mapped is not None
        x_expected = SimpleNamespace(
            num=3,
            num_squared=9,
            double_num_squared=18,
            double_num_squared_plus_1=19,
            text="Henry has fun",
        )
        self.assertEqual(x_mapped, x_expected)

    def test_mapper_pre_decorator(self) -> None:
        @lambda_mapper()
        def mapper_no_pre(x: DataPoint) -> DataPoint:
            x.double_num_squared = 2 * x.num_squared
            return x

        x = self._get_x(3)
        with self.assertRaises(AttributeError):
            x_mapped = mapper_no_pre(x)

        @lambda_mapper(pre=[square])
        def mapper_pre(x: DataPoint) -> DataPoint:
            x.double_num_squared = 2 * x.num_squared
            return x

        x = self._get_x(3)
        x_mapped = mapper_pre(x)
        self.assertEqual(x, self._get_x(3))
        assert x_mapped is not None
        x_expected = SimpleNamespace(
            num=3, num_squared=9, double_num_squared=18, text="Henry has fun"
        )
        self.assertEqual(x_mapped, x_expected)

        @lambda_mapper(pre=[mapper_pre])
        def mapper_pre_2(x: DataPoint) -> DataPoint:
            x.double_num_squared_plus_1 = x.double_num_squared + 1
            return x

        x = self._get_x(3)
        x_mapped = mapper_pre_2(x)
        self.assertEqual(x, self._get_x(3))
        assert x_mapped is not None
        x_expected = SimpleNamespace(
            num=3,
            num_squared=9,
            double_num_squared=18,
            double_num_squared_plus_1=19,
            text="Henry has fun",
        )
        self.assertEqual(x_mapped, x_expected)

    def test_decorator_mapper_memoized(self) -> None:
        square_hit_tracker = SquareHitTracker()

        @lambda_mapper(memoize=True)
        def square(x: DataPoint) -> DataPoint:
            x.num_squared = square_hit_tracker(x.num)
            return x

        x8 = self._get_x()
        x9 = self._get_x(9)
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 1)
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 1)
        x9_mapped = square(x9)
        assert x9_mapped is not None
        self.assertEqual(x9_mapped.num_squared, 81)
        self.assertEqual(square_hit_tracker.n_hits, 2)
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 2)

        square.reset_cache()
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 3)

    def test_decorator_mapper_memoized_none(self) -> None:
        square_hit_tracker = SquareHitTracker()

        @lambda_mapper(memoize=True)
        def square(x: DataPoint) -> DataPoint:
            x.num_squared = square_hit_tracker(x.num)
            if x.num == 21:
                return None
            return x

        x21 = self._get_x(21)
        x21_mapped = square(x21)
        self.assertIsNone(x21_mapped)
        self.assertEqual(square_hit_tracker.n_hits, 1)
        x21_mapped = square(x21)
        self.assertIsNone(x21_mapped)
        self.assertEqual(square_hit_tracker.n_hits, 1)

    def test_decorator_mapper_memoized_use_memoize_key(self) -> None:
        square_hit_tracker = SquareHitTracker()

        @lambda_mapper(memoize=True, memoize_key=lambda x: x.uid)
        def square(x: DataPoint) -> DataPoint:
            x.num_squared = square_hit_tracker(x.num)
            return x

        x1 = SimpleNamespace(
            uid="id1", num=8, not_used=0, unhashable=pd.DataFrame({"value": [5]})
        )
        x1_mapped = square(x1)
        assert x1_mapped is not None
        self.assertEqual(x1_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 1)
        x2 = SimpleNamespace(
            uid="id1", num=8, not_used=1, unhashable=pd.DataFrame({"value": [5]})
        )
        x2_mapped = square(x2)
        assert x2_mapped is not None
        self.assertEqual(x2_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 1)

    def test_decorator_mapper_not_memoized(self) -> None:
        square_hit_tracker = SquareHitTracker()

        @lambda_mapper(memoize=False)
        def square(x: DataPoint) -> DataPoint:
            x.num_squared = square_hit_tracker(x.num)
            return x

        x8 = self._get_x()
        x9 = self._get_x(9)
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 1)
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 2)
        x19_mapped = square(x9)
        assert x19_mapped is not None
        self.assertEqual(x19_mapped.num_squared, 81)
        self.assertEqual(square_hit_tracker.n_hits, 3)
        x8_mapped = square(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.num_squared, 64)
        self.assertEqual(square_hit_tracker.n_hits, 4)

    def test_mapper_pre_memoized(self) -> None:
        square_hit_tracker = SquareHitTracker()

        @lambda_mapper(memoize=False)
        def square(x: DataPoint) -> DataPoint:
            x.num_squared = square_hit_tracker(x.num)
            return x

        @lambda_mapper(pre=[square], memoize=True)
        def mapper_pre(x: DataPoint) -> DataPoint:
            x.double_num_squared = 2 * x.num_squared
            return x

        x8 = self._get_x()
        x9 = self._get_x(9)
        x8_mapped = mapper_pre(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.double_num_squared, 128)
        self.assertEqual(square_hit_tracker.n_hits, 1)
        x8_mapped = mapper_pre(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.double_num_squared, 128)
        self.assertEqual(square_hit_tracker.n_hits, 1)
        x9_mapped = mapper_pre(x9)
        assert x9_mapped is not None
        self.assertEqual(x9_mapped.double_num_squared, 162)
        self.assertEqual(square_hit_tracker.n_hits, 2)
        x8_mapped = mapper_pre(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.double_num_squared, 128)
        self.assertEqual(square_hit_tracker.n_hits, 2)

        mapper_pre.reset_cache()
        x8_mapped = mapper_pre(x8)
        assert x8_mapped is not None
        self.assertEqual(x8_mapped.double_num_squared, 128)
        self.assertEqual(square_hit_tracker.n_hits, 3)

    def test_mapper_decorator_no_parens(self) -> None:
        with self.assertRaisesRegex(ValueError, "missing parentheses"):

            @lambda_mapper
            def square(x: DataPoint) -> DataPoint:
                x.num_squared = x.num**2
                return x

    def test_mapper_with_args_kwargs(self) -> None:
        with self.assertRaises(ValueError):
            MapperWithArgs("my_mapper")

        with self.assertRaises(ValueError):
            MapperWithKwargs("my_mapper")


class TestGetHashable(unittest.TestCase):
    def test_get_hashable_hashable(self) -> None:
        x = (8, "abc")
        x_hashable = get_hashable(x)
        self.assertEqual(x, x_hashable)

    def test_get_hashable_dict(self) -> None:
        d = dict(a=8, b=dict(c=9, d="foo"))
        d_hashable = get_hashable(d)
        d_sub_expected = frozenset((("c", 9), ("d", "foo")))
        d_expected = frozenset((("a", 8), ("b", d_sub_expected)))
        self.assertEqual(d_hashable, d_expected)
        self.assertEqual(hash(d_hashable), hash(d_expected))

    def test_get_hashable_list(self) -> None:
        c = [8, dict(c=9, d="foo")]
        c_hashable = get_hashable(c)
        c_expected = (8, frozenset((("c", 9), ("d", "foo"))))
        self.assertEqual(c_hashable, c_expected)
        self.assertEqual(hash(c_hashable), hash(c_expected))

    def test_get_hashable_series(self) -> None:
        s = pd.Series(dict(a=8, b=dict(c=9, d="foo")), name="bar")
        s_hashable = get_hashable(s)
        s_sub_expected = frozenset((("c", 9), ("d", "foo")))
        s_expected = frozenset((("a", 8), ("b", s_sub_expected)))
        self.assertEqual(s_hashable, s_expected)
        self.assertEqual(hash(s_hashable), hash(s_expected))

    def test_get_hashable_series_with_doc(self) -> None:
        nlp = spacy.load("en_core_web_sm")
        doc = nlp("Foo went to the bar.")
        s = pd.Series(dict(a=8, b=doc), name="baz")
        s_hashable = get_hashable(s)
        s_expected = frozenset((("a", 8), ("b", doc)))
        self.assertEqual(s_hashable, s_expected)
        self.assertEqual(hash(s_hashable), hash(s_expected))

    def test_get_hashable_ndarray(self) -> None:
        v = np.array([[3, 6, 9], [0.4, 0.8, 0.12]])
        x = (8, dict(a=v))
        x_hashable = get_hashable(x)
        x_expected = (8, frozenset((("a", v.data.tobytes()),)))
        self.assertEqual(x_hashable, x_expected)

    def test_get_hashable_unhashable(self) -> None:
        v = pd.DataFrame(dict(a=[4, 5], b=[1, 2]))
        x = (8, dict(a=v))
        with self.assertRaises(ValueError):
            get_hashable(x)



import unittest

import torch

from snorkel.classification import DictDataLoader, DictDataset
from snorkel.classification.data import (
    DEFAULT_DATASET_NAME,
    DEFAULT_INPUT_DATA_KEY,
    DEFAULT_TASK_NAME,
)


class DatasetTest(unittest.TestCase):
    def test_classifier_dataset(self):
        """Unit test of DictDataset"""

        x1 = [
            torch.Tensor([1]),
            torch.Tensor([1, 2]),
            torch.Tensor([1, 2, 3]),
            torch.Tensor([1, 2, 3, 4]),
            torch.Tensor([1, 2, 3, 4, 5]),
        ]

        y1 = torch.Tensor([0, 0, 0, 0, 0])

        dataset = DictDataset(
            X_dict={"data1": x1}, Y_dict={"task1": y1}, name="new_data", split="train"
        )

        # Check if the dataset is correctly constructed
        self.assertTrue(torch.equal(dataset[0][0]["data1"], x1[0]))
        self.assertTrue(torch.equal(dataset[0][1]["task1"], y1[0]))
        self.assertEqual(
            repr(dataset),
            "DictDataset(name=new_data, X_keys=['data1'], Y_keys=['task1'])",
        )

        # Test from_tensors inits with default values
        dataset = DictDataset.from_tensors(x1, y1, "train")
        self.assertEqual(
            repr(dataset),
            f"DictDataset(name={DEFAULT_DATASET_NAME}, "
            f"X_keys=['{DEFAULT_INPUT_DATA_KEY}'], Y_keys=['{DEFAULT_TASK_NAME}'])",
        )

    def test_classifier_dataloader(self):
        """Unit test of DictDataLoader"""

        x1 = [
            torch.Tensor([1]),
            torch.Tensor([1, 2]),
            torch.Tensor([1, 2, 3]),
            torch.Tensor([1, 2, 3, 4]),
            torch.Tensor([1, 2, 3, 4, 5]),
        ]

        y1 = torch.Tensor([0, 0, 0, 0, 0])

        x2 = [
            torch.Tensor([1, 2, 3, 4, 5]),
            torch.Tensor([1, 2, 3, 4]),
            torch.Tensor([1, 2, 3]),
            torch.Tensor([1, 2]),
            torch.Tensor([1]),
        ]

        y2 = torch.Tensor([1, 1, 1, 1, 1])

        dataset = DictDataset(
            name="new_data",
            split="train",
            X_dict={"data1": x1, "data2": x2},
            Y_dict={"task1": y1, "task2": y2},
        )

        dataloader1 = DictDataLoader(dataset=dataset, batch_size=2)

        x_batch, y_batch = next(iter(dataloader1))

        # Check if the dataloader is correctly constructed
        self.assertEqual(dataloader1.dataset.split, "train")
        self.assertTrue(torch.equal(x_batch["data1"], torch.Tensor([[1, 0], [1, 2]])))
        self.assertTrue(
            torch.equal(
                x_batch["data2"], torch.Tensor([[1, 2, 3, 4, 5], [1, 2, 3, 4, 0]])
            )
        )
        self.assertTrue(torch.equal(y_batch["task1"], torch.Tensor([0, 0])))
        self.assertTrue(torch.equal(y_batch["task2"], torch.Tensor([1, 1])))

        dataloader2 = DictDataLoader(dataset=dataset, batch_size=3)

        x_batch, y_batch = next(iter(dataloader2))

        # Check if the dataloader with differet batch size is correctly constructed
        self.assertEqual(dataloader2.dataset.split, "train")
        self.assertTrue(
            torch.equal(
                x_batch["data1"], torch.Tensor([[1, 0, 0], [1, 2, 0], [1, 2, 3]])
            )
        )
        self.assertTrue(
            torch.equal(
                x_batch["data2"],
                torch.Tensor([[1, 2, 3, 4, 5], [1, 2, 3, 4, 0], [1, 2, 3, 0, 0]]),
            )
        )
        self.assertTrue(torch.equal(y_batch["task1"], torch.Tensor([0, 0, 0])))
        self.assertTrue(torch.equal(y_batch["task2"], torch.Tensor([1, 1, 1])))

        y3 = [
            torch.Tensor([2]),
            torch.Tensor([2]),
            torch.Tensor([2]),
            torch.Tensor([2]),
            torch.Tensor([2]),
        ]

        dataset.Y_dict["task2"] = y3

        x_batch, y_batch = next(iter(dataloader1))
        # Check dataloader is correctly updated with update dataset
        self.assertTrue(
            torch.equal(
                x_batch["data2"], torch.Tensor([[1, 2, 3, 4, 5], [1, 2, 3, 4, 0]])
            )
        )
        self.assertTrue(torch.equal(y_batch["task2"], torch.Tensor([[2], [2]])))

        x_batch, y_batch = next(iter(dataloader2))
        self.assertTrue(
            torch.equal(
                x_batch["data2"],
                torch.Tensor([[1, 2, 3, 4, 5], [1, 2, 3, 4, 0], [1, 2, 3, 0, 0]]),
            )
        )
        self.assertTrue(torch.equal(y_batch["task2"], torch.Tensor([[2], [2], [2]])))


if __name__ == "__main__":
    unittest.main()



import unittest
from types import SimpleNamespace

from snorkel.preprocess.nlp import SpacyPreprocessor


class TestSpacyPreprocessor(unittest.TestCase):
    def test_spacy_preprocessor(self) -> None:
        x = SimpleNamespace(text="Jane plays soccer.")
        preprocessor = SpacyPreprocessor("text", "doc")
        x_preprocessed = preprocessor(x)
        assert x_preprocessed is not None
        self.assertEqual(len(x_preprocessed.doc), 4)
        token = x_preprocessed.doc[0]
        self.assertEqual(token.text, "Jane")
        self.assertEqual(token.pos_, "PROPN")



from snorkel.preprocess.nlp import SpacyPreprocessor
from snorkel.preprocess.spark import make_spark_preprocessor

from .nlp import (
    BaseNLPLabelingFunction,
    SpacyPreprocessorParameters,
    base_nlp_labeling_function,
)


class SparkNLPLabelingFunction(BaseNLPLabelingFunction):
    r"""Special labeling function type for SpaCy-based LFs running on Spark.

    This class is a Spark-compatible version of ``NLPLabelingFunction``.
    See ``NLPLabelingFunction`` for details.

    Parameters
    ----------
    name
        Name of the LF
    f
        Function that implements the core LF logic
    resources
        Labeling resources passed in to ``f`` via ``kwargs``
    pre
        Preprocessors to run before SpacyPreprocessor is executed
    text_field
        Name of data point text field to input
    doc_field
        Name of data point field to output parsed document to
    language
        SpaCy model to load
        See https://spacy.io/usage/models#usage
    disable
        List of pipeline components to disable
        See https://spacy.io/usage/processing-pipelines#disabling
    memoize
        Memoize preprocessor outputs?
    memoize_key
        Hashing function to handle the memoization (default to snorkel.map.core.get_hashable)
    gpu
        Prefer Spacy GPU processing?

    Raises
    ------
    ValueError
        Calling incorrectly defined preprocessors

    Attributes
    ----------
    name
        See above
    """

    @classmethod
    def _create_preprocessor(
        cls, parameters: SpacyPreprocessorParameters
    ) -> SpacyPreprocessor:
        preprocessor = SpacyPreprocessor(**parameters._asdict())
        make_spark_preprocessor(preprocessor)
        return preprocessor


class spark_nlp_labeling_function(base_nlp_labeling_function):
    """Decorator to define a SparkNLPLabelingFunction object from a function.

    Parameters
    ----------
    name
        Name of the LF
    resources
        Labeling resources passed in to ``f`` via ``kwargs``
    pre
        Preprocessors to run before SpacyPreprocessor is executed
    text_field
        Name of data point text field to input
    doc_field
        Name of data point field to output parsed document to
    language
        SpaCy model to load
        See https://spacy.io/usage/models#usage
    disable
        List of pipeline components to disable
        See https://spacy.io/usage/processing-pipelines#disabling
    memoize
        Memoize preprocessor outputs?
    memoize_key
        Hashing function to handle the memoization (default to snorkel.map.core.get_hashable)

    Example
    -------
    >>> @spark_nlp_labeling_function()
    ... def has_person_mention(x):
    ...     person_ents = [ent for ent in x.doc.ents if ent.label_ == "PERSON"]
    ...     return 0 if len(person_ents) > 0 else -1
    >>> has_person_mention
    SparkNLPLabelingFunction has_person_mention, Preprocessors: [SpacyPreprocessor...]

    >>> from pyspark.sql import Row
    >>> x = Row(text="The movie was good.")
    >>> has_person_mention(x)
    -1
    """

    _lf_cls = SparkNLPLabelingFunction



from typing import Dict

from setuptools import find_packages, setup

# version.py defines the VERSION and VERSION_SHORT variables.
# We use exec here so we don't import snorkel.
VERSION: Dict[str, str] = {}
with open("snorkel/version.py", "r") as version_file:
    exec(version_file.read(), VERSION)

# Use README.md as the long_description for the package
with open("README.md", "r") as readme_file:
    long_description = readme_file.read()

setup(
    name="snorkel",
    version=VERSION["VERSION"],
    url="https://github.com/snorkel-team/snorkel",
    description="A system for quickly generating training data with weak supervision",
    long_description_content_type="text/markdown",
    long_description=long_description,
    license="Apache License 2.0",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
    ],
    project_urls={
        "Homepage": "https://snorkel.org",
        "Source": "https://github.com/snorkel-team/snorkel/",
        "Bug Reports": "https://github.com/snorkel-team/snorkel/issues",
        "Citation": "https://doi.org/10.14778/3157794.3157797",
    },
    packages=find_packages(exclude=("test*",)),
    include_package_data=True,
    install_requires=[
        "munkres>=1.0.6",
        "numpy>=1.24.0",
        "scipy>=1.2.0",
        "pandas>=1.0.0",
        "tqdm>=4.33.0",
        "scikit-learn>=0.20.2",
        "torch>=1.2.0",
        "tensorboard>=2.13.0",
        "protobuf>=3.19.6",
        "networkx>=2.2",
    ],
    python_requires=">=3.11",
    keywords="machine-learning ai weak-supervision",
)



import unittest
from types import SimpleNamespace

from snorkel.slicing import SlicingFunction, slicing_function


class TestSlicingFunction(unittest.TestCase):
    def _run_sf(self, sf: SlicingFunction) -> None:
        x_43 = SimpleNamespace(num=43)
        x_19 = SimpleNamespace(num=19)
        self.assertEqual(sf(x_43), True)
        self.assertEqual(sf(x_19), False)

    def _run_sf_raise(self, sf: SlicingFunction) -> None:
        x_none = SimpleNamespace(num=None)
        with self.assertRaises(TypeError):
            sf(x_none)

    def test_slicing_function_decorator(self) -> None:
        @slicing_function()
        def sf(x) -> int:
            return x.num > 42

        self.assertIsInstance(sf, SlicingFunction)
        self.assertEqual(sf.name, "sf")
        self._run_sf(sf)
        self._run_sf_raise(sf)

    def test_slicing_function_decorator_no_parens(self) -> None:
        with self.assertRaisesRegex(ValueError, "missing parentheses"):

            @slicing_function
            def sf(x) -> int:
                return 0 if x.num > 42 else -1
